# -*- coding: utf-8 -*-
"""Forcing schemes (:mod:`fluidsim.base.forcing.specific`)
================================================================

Provides:

.. autoclass:: SpecificForcing
   :members:
   :private-members:

.. autoclass:: SpecificForcingPseudoSpectralSimple
   :members:
   :private-members:

.. autoclass:: InScriptForcingPseudoSpectral
   :members:
   :private-members:

.. autoclass:: SpecificForcingPseudoSpectralCoarse
   :members:
   :private-members:

.. autoclass:: InScriptForcingPseudoSpectralCoarse
   :members:
   :private-members:

.. autoclass:: NormalizedForcing
   :members:
   :private-members:

.. autoclass:: Proportional
   :members:
   :private-members:

.. autoclass:: RandomSimplePseudoSpectral
   :members:
   :private-members:

.. autoclass:: TimeCorrelatedRandomPseudoSpectral
   :members:
   :private-members:

"""

from copy import deepcopy
from math import radians
import types
from warnings import warn
from pathlib import Path

import numpy as np

from fluiddyn.util import mpi
from fluiddyn.calcul.easypyfft import fftw_grid_size

from fluidsim.base.setofvariables import SetOfVariables


class SpecificForcing:
    """Base class for specific forcing"""

    tag = "specific"

    @classmethod
    def _complete_params_with_default(cls, params):
        params.forcing.available_types.append(cls.tag)

    def __init__(self, sim):

        self.sim = sim
        self.oper = sim.oper
        self.params = sim.params


class SpecificForcingPseudoSpectralSimple(SpecificForcing):
    """Specific forcing for pseudo-spectra solvers"""

    tag = "pseudo_spectral"

    def __init__(self, sim):
        super().__init__(sim)
        self.fstate = sim.state.__class__(sim, oper=self.sim.oper)
        self.forcing_fft = self.fstate.state_spect

    def compute(self):
        """compute the forcing."""
        obj = self.compute_forcing_fft_each_time()
        if isinstance(obj, dict):
            kwargs = obj
        else:
            if self.sim.params.forcing.key_forced is None:
                raise ValueError("params.forcing.key_forced must be initialized.")
            kwargs = {self.sim.params.forcing.key_forced: obj}
        self.fstate.init_statespect_from(**kwargs)

    def compute_forcing_fft_each_time(self):
        raise NotImplementedError


class InScriptForcingPseudoSpectral(SpecificForcingPseudoSpectralSimple):
    """Forcing maker for forcing defined by the user in the launching script

    .. inheritance-diagram:: InScriptForcingPseudoSpectral

    """

    tag = "in_script"

    def __init__(self, sim):
        super().__init__(sim)
        self.is_initialized = False

    def compute_forcing_fft_each_time(self):
        """Compute the coarse forcing in Fourier space"""
        obj = self.compute_forcing_each_time()
        if isinstance(obj, dict):
            kwargs = {key: self.sim.oper.fft(value) for key, value in obj.items()}
        else:
            if self.sim.params.forcing.key_forced is None:
                raise ValueError("params.forcing.key_forced must be initialized.")
            kwargs = {self.sim.params.forcing.key_forced: self.sim.oper.fft(obj)}
        return kwargs

    def compute_forcing_each_time(self):
        """Compute the coarse forcing in real space"""
        return self.sim.oper.create_arrayX(value=0)

    def monkeypatch_compute_forcing_fft_each_time(self, func):
        """Replace the method by a user-defined method"""
        self.compute_forcing_fft_each_time = types.MethodType(func, self)
        self.is_initialized = True

    def monkeypatch_compute_forcing_each_time(self, func):
        """Replace the method by a user-defined method"""
        self.compute_forcing_each_time = types.MethodType(func, self)
        self.is_initialized = True


class SpecificForcingPseudoSpectralCoarse(SpecificForcing):
    """Specific forcing for pseudo-spectra solvers"""

    tag = "pseudo_spectral"
    _key_forced_default = "rot_fft"

    @staticmethod
    def _check_forcing_shape(shape_forcing, shape):
        """Check if shape of the forcing array exceeds the shape
        of the global array.

        Parameters
        ----------
        shape_forcing: array-like
            A single-element array containing index of largest forcing
            wavenumber or a tuple indincating shape of the forcing array.

        shape: array-like
            A tuple indicating the shape of an array or Operators instance.

        """
        if any(np.greater(shape_forcing, shape)):
            raise NotImplementedError(
                "The resolution is too small for the required forcing: "
                f"any(np.greater({shape_forcing}, {shape}))"
            )

    def __init__(self, sim):

        super().__init__(sim)

        params = sim.params

        self.forcing_fft = SetOfVariables(
            like=sim.state.state_spect, info="forcing_fft", value=0.0
        )

        if params.forcing.nkmax_forcing < params.forcing.nkmin_forcing:
            raise ValueError(
                "params.forcing.nkmax_forcing < \n" "params.forcing.nkmin_forcing"
            )

        self.kmax_forcing = self.oper.deltak * params.forcing.nkmax_forcing
        self.kmin_forcing = self.oper.deltak * params.forcing.nkmin_forcing

        self.forcing_rate = params.forcing.forcing_rate

        if params.forcing.key_forced is not None:
            self.key_forced = params.forcing.key_forced
        else:
            self.key_forced = self._key_forced_default

        try:
            n = 2 * fftw_grid_size(params.forcing.nkmax_forcing)
        except ImportError:
            warn("To use smaller forcing arrays: pip install pulp")
            i = 0
            while 2 * params.forcing.nkmax_forcing > 2 ** i:
                i += 1
            n = 2 ** i

        self._check_forcing_shape([n], sim.oper.shapeX_seq)

        try:
            angle = self.params.forcing[self.tag].angle
        except AttributeError:
            pass
        else:
            if isinstance(angle, str):
                if angle.endswith("°"):
                    angle = radians(float(angle[:-1]))
                else:
                    raise ValueError(
                        "Angle should be a string with \n"
                        + "the degree symbol or a float in radians"
                    )

            self.angle = angle
            self.kxmax_forcing = np.sin(angle) * self.kmax_forcing
            self.kymax_forcing = np.cos(angle) * self.kmax_forcing

        if mpi.rank == 0:
            params_coarse = deepcopy(params)

            params_coarse.oper.nx = n
            # The "+ 1" aims to give some gap between the kxmax and
            # the boundary of the oper_coarse.
            try:
                params_coarse.oper.nx = 2 * fftw_grid_size(
                    int(self.kxmax_forcing / self.oper.deltakx) + 1
                )
            except AttributeError:
                pass

            try:
                params_coarse.oper.ny = n
                try:
                    params_coarse.oper.ny = 2 * fftw_grid_size(
                        int(self.kymax_forcing / self.oper.deltaky) + 1
                    )
                except AttributeError:
                    pass
            except AttributeError:
                pass

            try:
                params_coarse.oper.nz = n
            except AttributeError:
                pass

            params_coarse.oper.type_fft = "sequential"
            params_coarse.oper.coef_dealiasing = 1.0

            self.oper_coarse = sim.oper.__class__(params=params_coarse)
            self.shapeK_loc_coarse = self.oper_coarse.shapeK_loc
            self.COND_NO_F = self._compute_cond_no_forcing()

            self.nb_forced_modes = (
                self.COND_NO_F.size
                - np.array(self.COND_NO_F, dtype=np.int32).sum()
            )
            if not self.nb_forced_modes:
                raise ValueError("0 modes forced.")

            try:
                hasattr(self, "plot_forcing_region")
            except NotImplementedError:
                pass
            else:
                mpi.printby0(
                    "To plot the forcing modes, you can use:\n"
                    "sim.forcing.forcing_maker.plot_forcing_region()"
                )

            self.ind_forcing = (
                np.logical_not(self.COND_NO_F).flatten().nonzero()[0]
            )

            self.fstate_coarse = sim.state.__class__(sim, oper=self.oper_coarse)
        else:
            self.shapeK_loc_coarse = None

        if mpi.nb_proc > 1:
            self.shapeK_loc_coarse = mpi.comm.bcast(
                self.shapeK_loc_coarse, root=0
            )

    def _compute_cond_no_forcing(self):
        if hasattr(self.oper_coarse, "K"):
            K = self.oper_coarse.K
        else:
            K = np.sqrt(self.oper_coarse.K2)
        COND_NO_F = np.logical_or(K > self.kmax_forcing, K < self.kmin_forcing)

        COND_NO_F[self.oper_coarse.shapeK_loc[0] // 2] = True
        COND_NO_F[:, self.oper_coarse.shapeK_loc[1] - 1] = True

        return COND_NO_F

    def put_forcingc_in_forcing(self):
        """Copy data from self.fstate_coarse.state_spect into forcing_fft."""
        if mpi.rank == 0:
            state_spect = self.fstate_coarse.state_spect
            oper_coarse = self.oper_coarse
        else:
            state_spect = None
            oper_coarse = None

        self.oper.put_coarse_array_in_array_fft(
            state_spect, self.forcing_fft, oper_coarse, self.shapeK_loc_coarse
        )

    def verify_injection_rate(self):
        """Verify injection rate."""
        f_fft = self.forcing_fft.get_var(self.key_forced)
        var_fft = self.sim.state.state_spect.get_var(self.key_forced)
        deltat = self.sim.time_stepping.deltat

        P_forcing1 = np.real(f_fft.conj() * var_fft)
        P_forcing2 = abs(f_fft) ** 2
        P_forcing2 = deltat / 2 * self.oper.sum_wavenumbers(P_forcing2)
        P_forcing1 = self.oper.sum_wavenumbers(P_forcing1)
        if mpi.rank == 0:
            print(
                "P_f = {:9.4e} ; P_1 = {:9.4e}; P_2 = {:9.4e}".format(
                    P_forcing1 + P_forcing2, P_forcing1, P_forcing2
                )
            )

    def verify_injection_rate_coarse(self, var_fft=None):
        """Verify injection rate."""
        if var_fft is None:
            var_fft = self.sim.state.state_spect.get_var(self.key_forced)
            var_fft = self.oper.coarse_seq_from_fft_loc(
                var_fft, self.shapeK_loc_coarse
            )

        if mpi.rank == 0:
            f_fft = self.fstate_coarse.get_var(self.key_forced)
            deltat = self.sim.time_stepping.deltat

            P_forcing1 = np.real(f_fft.conj() * var_fft)
            P_forcing2 = abs(f_fft) ** 2
            P_forcing2 = deltat / 2 * self.oper_coarse.sum_wavenumbers(P_forcing2)
            P_forcing1 = self.oper_coarse.sum_wavenumbers(P_forcing1)
            print(
                "P_f = {:9.4e} ; P_1 = {:9.4e}; P_2 = {:9.4e} (coarse)".format(
                    P_forcing1 + P_forcing2, P_forcing1, P_forcing2
                )
            )

    def compute(self):
        """compute a forcing normalize with a 2nd degree eq."""

        if mpi.rank == 0:
            obj = self.compute_forcingc_fft_each_time()
            if isinstance(obj, dict):
                kwargs = obj
            else:
                kwargs = {self.key_forced: obj}
            self.fstate_coarse.init_statespect_from(**kwargs)
            self.oper_coarse.dealiasing_setofvar(self.fstate_coarse.state_spect)

        self.put_forcingc_in_forcing()

    def compute_forcingc_fft_each_time(self):
        raise NotImplementedError


class InScriptForcingPseudoSpectralCoarse(SpecificForcingPseudoSpectralCoarse):
    """Forcing maker for forcing defined by the user in the launching script

    .. inheritance-diagram:: InScriptForcingPseudoSpectralCoarse

    """

    tag = "in_script_coarse"

    def __init__(self, sim):
        super().__init__(sim)
        self.is_initialized = False

    def compute_forcingc_fft_each_time(self):
        """Compute the coarse forcing in Fourier space"""
        return self.oper_coarse.fft(self.compute_forcingc_each_time())

    def compute_forcingc_each_time(self):
        """Compute the coarse forcing in real space"""
        return self.oper_coarse.create_arrayX(value=0)

    def monkeypatch_compute_forcingc_fft_each_time(self, func):
        """Replace the method by a user-defined method"""
        self.compute_forcingc_fft_each_time = types.MethodType(func, self)
        self.is_initialized = True

    def monkeypatch_compute_forcingc_each_time(self, func):
        """Replace the method by a user-defined method"""
        self.compute_forcingc_each_time = types.MethodType(func, self)
        self.is_initialized = True


class Proportional(SpecificForcingPseudoSpectralCoarse):
    """Specific forcing proportional to the forced variable

    .. inheritance-diagram:: Proportional

    """

    tag = "proportional"

    def compute(self):
        """compute a forcing normalize with a 2nd degree eq."""

        try:
            a_fft = self.sim.state.state_spect.get_var(self.key_forced)
        except ValueError:
            a_fft = self.sim.state.get_var(self.key_forced)

        a_fft = self.oper.coarse_seq_from_fft_loc(a_fft, self.shapeK_loc_coarse)

        if mpi.rank == 0:
            fa_fft = self.forcingc_raw_each_time(a_fft)
            kwargs = {self.key_forced: fa_fft}
            self.fstate_coarse.init_statespect_from(**kwargs)

        self.put_forcingc_in_forcing()

    def forcingc_raw_each_time(self, vc_fft):
        """Modify the array fvc_fft to fixe the injection rate.

        varc : ndarray
            a variable at the coarse resolution.

        To be called only with proc 0.
        """
        fvc_fft = vc_fft.copy()
        fvc_fft[self.COND_NO_F] = 0.0

        Z_fft = abs(fvc_fft) ** 2 / 2.0

        # # possibly "kill" the largest mode
        # nb_kill = 0
        # for ik in xrange(nb_kill):
        #     imax = Z_fft.argmax()
        #     Z_fft.flat[imax] = 0.
        #     fvc_fft.flat[imax] = 0.

        # # possibly add randomness: random kill!
        # nb_kill = self.nb_forced_modes-10
        # ind_kill = random.sample(self.ind_forcing,nb_kill)
        # for ik in ind_kill:
        #     Z_fft.flat[ik] = 0.
        #     fvc_fft.flat[ik] = 0.

        Z = self.oper_coarse.sum_wavenumbers(Z_fft)
        deltat = self.sim.time_stepping.deltat
        alpha = (np.sqrt(1 + deltat * self.forcing_rate / Z) - 1.0) / deltat
        fvc_fft = alpha * fvc_fft

        return fvc_fft


class NormalizedForcing(SpecificForcingPseudoSpectralCoarse):
    """Specific forcing normalized to keep constant injection

    .. inheritance-diagram:: NormalizedForcing

    """

    tag = "normalized"

    @classmethod
    def _complete_params_with_default(cls, params):
        """This static method is used to complete the *params* container."""
        super()._complete_params_with_default(params)
        try:
            params.forcing.normalized
        except AttributeError:
            params.forcing._set_child(
                "normalized",
                {
                    "type": "2nd_degree_eq",
                    "which_root": "minabs",
                    "constant_rate_of": None,
                },
            )

            params.forcing._set_doc("How the forcing is normalized")

    def __init__(self, sim):

        super().__init__(sim)

        params_norm = self.params.forcing.normalized

        if not hasattr(params_norm, "constant_rate_of"):
            params_norm._set_attr("constant_rate_of", None)

        if (
            params_norm.constant_rate_of is not None
            and params_norm.type != "2nd_degree_eq"
        ):
            raise NotImplementedError(
                "params.forcing.normalized.constant_rate_of is implemented "
                'only for params.forcing.normalized.type == "2nd_degree_eq"'
            )

    def compute(self):
        """compute a forcing normalize with a 2nd degree eq."""

        try:
            a_fft = self.sim.state.state_spect.get_var(self.key_forced)
        except ValueError:
            a_fft = self.sim.state.get_var(self.key_forced)

        try:
            a_fft = self.oper.coarse_seq_from_fft_loc(
                a_fft, self.shapeK_loc_coarse
            )
        except IndexError:
            raise ValueError(
                "rank={}, shapeK_loc(coarse)={}, shapeK_loc={}".format(
                    self.oper.rank, self.shapeK_loc_coarse, self.oper.shapeK_loc
                )
            )

        if mpi.rank == 0:
            fa_fft = self.forcingc_raw_each_time(a_fft)
            fa_fft = self.normalize_forcingc(fa_fft, a_fft)
            kwargs = {self.key_forced: fa_fft}
            self.fstate_coarse.init_statespect_from(**kwargs)
        # print('check forcing injection')
        # self.verify_injection_rate_coarse(a_fft)

        self.put_forcingc_in_forcing()

        # # check...
        # f_fft = self.forcing_fft.get_var(self.key_forced)
        # fc_fft = self.oper.coarse_seq_from_fft_loc(
        #     f_fft, self.shapeK_loc_coarse)

        # if mpi.rank == 0:
        #     assert np.allclose(fa_fft, fc_fft)

        # # verification
        # self.verify_injection_rate()

        if mpi.rank == 0:
            return fa_fft

    def normalize_forcingc(self, fvc_fft, vc_fft):
        """Normalize the coarse forcing"""

        type_normalize = self.params.forcing.normalized.type

        if type_normalize == "2nd_degree_eq":
            return self.normalize_forcingc_2nd_degree_eq(fvc_fft, vc_fft)

        elif type_normalize == "particular_k":
            return self.normalize_forcingc_part_k(fvc_fft, vc_fft)

        else:
            ValueError(
                "Bad value for parameter forcing.type_normalize:", type_normalize
            )

    def normalize_forcingc_part_k(self, fvc_fft, vc_fft):
        """Modify the array fvc_fft to fixe the injection rate.

        To be called only with proc 0.

        Parameters
        ----------

        fvc_fft : ndarray
            The non-normalized forcing at the coarse resolution.

        vc_fft : ndarray
            The forced variable at the coarse resolution.

        """
        oper_c = self.oper_coarse

        oper_c.project_fft_on_realX(fvc_fft)
        # fvc_fft[self.COND_NO_F] = 0.

        P_forcing2 = np.real(fvc_fft.conj() * vc_fft)
        P_forcing2 = oper_c.sum_wavenumbers(P_forcing2)

        # we choice randomly a "particular" wavenumber
        # in the forced space
        KX_f = oper_c.KX[~self.COND_NO_F].flatten()
        KY_f = oper_c.KY[~self.COND_NO_F].flatten()
        nb_wn_f = len(KX_f)

        # warning : this is 2d specific!

        ipart = np.random.randint(0, nb_wn_f - 1)
        kx_part = KX_f[ipart]
        ky_part = KY_f[ipart]
        ikx_part = abs(oper_c.kx_loc - kx_part).argmin()
        iky_part = abs(oper_c.ky_loc - ky_part).argmin()

        ik0_part = iky_part
        ik1_part = ikx_part

        P_forcing2_part = np.real(
            fvc_fft[ik0_part, ik1_part].conj() * vc_fft[ik0_part, ik1_part]
            + fvc_fft[ik0_part, ik1_part] * vc_fft[ik0_part, ik1_part].conj()
        )

        if ikx_part == 0:
            P_forcing2_part = P_forcing2_part / 2.0
        P_forcing2_other = P_forcing2 - P_forcing2_part
        fvc_fft[ik0_part, ik1_part] = (
            -P_forcing2_other / vc_fft[ik0_part, ik1_part].real
        )

        if ikx_part != 0:
            fvc_fft[ik0_part, ik1_part] = fvc_fft[ik0_part, ik1_part] / 2.0

        oper_c.project_fft_on_realX(fvc_fft)

        # normalisation to obtain the wanted total forcing rate
        PZ_nonorm = (
            oper_c.sum_wavenumbers(abs(fvc_fft) ** 2)
            * self.sim.time_stepping.deltat
            / 2
        )
        fvc_fft = fvc_fft * np.sqrt(float(self.forcing_rate) / PZ_nonorm)

        return fvc_fft

    def normalize_forcingc_2nd_degree_eq(self, fvc_fft, vc_fft):
        r"""Modify the array fvc_fft to fixe the injection rate.

        To be called only with proc 0.

        .. |p| mathmacro:: \partial

        .. |var| mathmacro:: \hat\alpha

        .. |fvar| mathmacro:: \hat f

        .. |Sum| mathmacro:: \sum_{\mathbf k}

        We consider that we force a variable |var| with a forcing |fvar|.

        .. math::

            \p_t \var = \fvar

        We want to normalize the forcing |fvar| such that the average over the
        time step of the injection of a quadratic quantity be equal to
        ``self.forcing_rate`` (:math:`P`). Let's consider that the time step
        starts at :math:`t=0` and that the time increment is :math:`\delta t`.

        For simplicity, we first consider that the quadratic quantity is the
        quadratic quantity of the forced variable |var|. Note that this
        function supports other quadratic quantities (for details, read the
        code). The average of the injection rate over the time step is:

        .. math::

            P = \int_0^{\delta t} \frac{dt}{\delta t} \Sum \p_t \frac{|\var^2|}{2}
            = \int_0^{\delta t} \frac{dt}{\delta t} \Sum \var^* \fvar

        We compute an approximation at first order in :math:`\delta t` so that
        we can normalize the forcing such that the value given by this
        approximation is constant for all time steps. For each time step, the
        forcing |fvar| is constant in time. At first order in :math:`\delta t`,
        we have during the time step:

        .. math::

            \var(t) \simeq \var(0) + \fvar t

        and we get

        .. math::

            P \simeq \fvar \int_0^{\delta t} \frac{dt}{\delta t} \Sum (\var(0)^* + \fvar^* t)
            = \Sum \var(0)^* \fvar + \Sum \frac{|\fvar|^2}{2} \delta t

        The final forcing |fvar| is proportional to the "random" forcing :math:`\fvar_r`:

        .. math:: \fvar = R \fvar_r

        We solve a second-order equation to get the value of the coefficient
        :math:`R`:

        .. math::

            \left(\Sum \frac{|\fvar_r|^2}{2} \delta t\right) R^2 + \left(\Sum \var(0)^* \fvar_r\right) R - P = 0

        Parameters
        ----------

        fvc_fft : ndarray
            The non-normalized forcing at the coarse resolution.

        vc_fft : ndarray
            The forced variable at the coarse resolution.
        """
        oper_c = self.oper_coarse

        deltat = self.sim.time_stepping.deltat

        if self.params.forcing.normalized.constant_rate_of is not None:
            if not hasattr(self.sim.forcing, "compute_coef_ab_normalize"):
                raise NotImplementedError
            a, b = self.sim.forcing.compute_coef_ab_normalize(
                self.params.forcing.normalized.constant_rate_of,
                self.key_forced,
                fvc_fft,
                vc_fft,
                deltat,
            )
        else:
            a = deltat / 2 * oper_c.sum_wavenumbers(abs(fvc_fft) ** 2)
            b = oper_c.sum_wavenumbers((vc_fft.conj() * fvc_fft).real)

        c = -self.forcing_rate

        alpha = self.coef_normalization_from_abc(a, b, c)

        fvc_fft = alpha * fvc_fft

        # P1 = deltat / 2 * oper_c.sum_wavenumbers(abs(fvc_fft)**2)
        # P2 = oper_c.sum_wavenumbers((vc_fft.conj() * fvc_fft).real)
        # print('forcing after normalization:', (P1 + P2)/self.forcing_rate,
        #       P1/self.forcing_rate)

        return fvc_fft

    def coef_normalization_from_abc(self, a, b, c):
        """Compute the roots of a quadratic equation

        Compute the roots given the coefficients ``a``, ``b`` and ``c``.
        Then, select one of the roots based on a criteria and return it.

        Notes
        -----
        Set params.forcing.normalized.which_root to choose the root with:

        - ``minabs`` : minimum absolute value
        - ``first`` : root with positive sign before discriminant
        - ``second`` : root with negative sign before discriminant
        - ``positive`` : positive root

        """
        try:
            alpha1, alpha2 = np.roots([a, b, c])
        except ValueError:
            return 0.0

        which_root = self.params.forcing.normalized.which_root

        if which_root == "minabs":
            if abs(alpha2) < abs(alpha1):
                return alpha2

            else:
                return alpha1

        elif which_root == "first":
            return alpha1

        elif which_root == "second":
            return alpha2

        elif which_root == "positive":
            if alpha2 > 0.0:
                return alpha2

            else:
                return alpha1

        else:
            raise ValueError(
                "Not sure how to choose which root to normalize forcing with."
            )


class RandomSimplePseudoSpectral(NormalizedForcing):
    """Random normalized forcing

    .. inheritance-diagram:: RandomSimplePseudoSpectral
    """

    tag = "random"

    @classmethod
    def _complete_params_with_default(cls, params):
        """This static method is used to complete the *params* container."""
        super()._complete_params_with_default(params)

        try:
            params.forcing.random
        except AttributeError:
            params.forcing._set_child("random", {"only_positive": False})

    def __init__(self, sim):

        super().__init__(sim)

        if self.params.forcing.random.only_positive:
            self._min_val = None
        else:
            self._min_val = -1

    def compute_forcingc_raw(self):
        """Random coarse forcing.

        To be called only with proc 0.
        """
        f_fft = self.oper_coarse.create_arrayK_random(min_val=self._min_val)
        # fftwpy/easypyfft returns f_fft
        f_fft[self.oper_coarse.shapeK_loc[0] // 2] = 0.0
        f_fft[:, self.oper_coarse.shapeK_loc[1] - 1] = 0.0
        f_fft = self.oper_coarse.project_fft_on_realX(f_fft)
        f_fft[self.COND_NO_F] = 0.0
        return f_fft

    def forcingc_raw_each_time(self, a_fft):
        return self.compute_forcingc_raw()


class TimeCorrelatedRandomPseudoSpectral(RandomSimplePseudoSpectral):
    """Time correlated random normalized forcing

    .. inheritance-diagram:: TimeCorrelatedRandomPseudoSpectral
    """

    tag = "tcrandom"

    @classmethod
    def _complete_params_with_default(cls, params):
        """This static method is used to complete the *params* container."""
        super()._complete_params_with_default(params)

        try:
            params.forcing.tcrandom
        except AttributeError:
            params.forcing._set_child(
                "tcrandom", {"time_correlation": "based_on_forcing_rate"}
            )

    def __init__(self, sim):

        super().__init__(sim)

        if mpi.rank == 0:

            self._forcing_state_file_path = (
                Path(sim.output.path_run) / "_forcing_state.txt"
            )

            if self._forcing_state_file_path.exists():
                with open(self._forcing_state_file_path) as file:
                    lines = file.readlines()

                t_last_change, seed0, seed1 = lines[-1].split()
                self.t_last_change = float(t_last_change)
                self._seed0 = int(seed0)
                self._seed1 = int(seed1)
            else:
                self.t_last_change = self.sim.time_stepping.t
                self._seed0 = np.random.randint(0, 2 ** 31)
                self._seed1 = np.random.randint(0, 2 ** 31)
                self._save_state()

            np.random.seed(self._seed0)
            self.forcing0 = self.compute_forcingc_raw()
            np.random.seed(self._seed1)
            self.forcing1 = self.compute_forcingc_raw()

            pforcing = self.params.forcing
            try:
                time_correlation = pforcing[self.tag].time_correlation
            except AttributeError:
                time_correlation = pforcing.tcrandom.time_correlation

            if time_correlation == "based_on_forcing_rate":
                self.period_change_f0f1 = self.forcing_rate ** (-1.0 / 3)
            else:
                self.period_change_f0f1 = time_correlation

    def forcingc_raw_each_time(self, a_fft):
        """Return a coarse forcing as a linear combination of 2 random arrays

        Compute the new random coarse forcing arrays when necessary.

        """
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_change >= self.period_change_f0f1:
            self.t_last_change = tsim
            self._seed0 = self._seed1
            self.forcing0 = self.forcing1
            self._seed1 = np.random.randint(0, 2 ** 31)
            self.forcing1 = self.compute_forcingc_raw()
            self._save_state()

        f_fft = self.forcingc_from_f0f1()
        return f_fft

    def _save_state(self):
        with open(self._forcing_state_file_path, "w") as file:
            file.write(
                "# do not modify by hand\n# t_last_change seed0 seed1\n"
                f"{self.t_last_change} {self._seed0} {self._seed1}\n"
            )

    def forcingc_from_f0f1(self):
        """Return a coarse forcing as a linear combination of 2 random arrays"""
        tsim = self.sim.time_stepping.t
        deltat = self.period_change_f0f1
        omega = np.pi / deltat

        deltaf = self.forcing1 - self.forcing0

        f_fft = (
            self.forcing1
            - 0.5 * (np.cos((tsim - self.t_last_change) * omega) + 1) * deltaf
        )

        return f_fft



import numpy as np

from fluiddyn.util import mpi

from .with_uxuy import Simul
from .solver import Simul as SimulBase

from .test_solver import TestSimulBase as Base


class TestSimulBase(Base):
    Simul = Simul


class TestSolverNS2DTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.output.HAS_TO_SAVE = False

    def test_tendency(self):
        sim = self.sim
        ux_fft = sim.state.get_var("ux_fft")
        uy_fft = sim.state.get_var("uy_fft")

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)
        Fx_fft = tend.get_var("ux_fft")
        Fy_fft = tend.get_var("uy_fft")

        T_rot = np.real(Fx_fft.conj() * ux_fft + Fy_fft.conj() * uy_fft)

        ratio = sim.oper.sum_wavenumbers(T_rot) / sim.oper.sum_wavenumbers(
            abs(T_rot)
        )

        self.assertGreater(1e-15, abs(ratio))


class TestForcingMilestone(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "milestone"
        movement = params.forcing.milestone.movement
        movement.type = "uniform"
        movement.uniform.speed = 1.0
        return params

    def test_milestone(self):
        self.sim.time_stepping.start()


class TestForcingMilestoneSinusoidal(TestForcingMilestone):
    Simul = SimulBase

    @classmethod
    def init_params(self):
        params = super().init_params()
        movement = params.forcing.milestone.movement
        movement.type = "sinusoidal"
        movement.sinusoidal.length = 6.0
        movement.sinusoidal.period = 100.0


class TestForcingMilestonePeriodicUniform(TestForcingMilestone):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.oper.NO_SHEAR_MODES = True
        params.time_stepping.t_end = 2.0
        params.forcing.milestone.nx_max = 16
        movement = params.forcing.milestone.movement
        movement.type = "periodic_uniform"
        movement.periodic_uniform.length = 2.0
        movement.periodic_uniform.length_acc = 0.25
        movement.periodic_uniform.speed = 2.5

    def test_milestone(self):
        super().test_milestone()

        if mpi.nb_proc > 1:
            return

        milestone = self.sim.forcing.forcing_maker
        print(milestone.period)
        milestone.check_plot_forcing(8.0)
        milestone.check_plot_solid(8.0)
        milestone.check_with_animation(number_frames=4, interval=1)



from time import time
import os
import sys
from datetime import timedelta

import numpy as np
import matplotlib.pyplot as plt

from fluiddyn.util import mpi, print_memory_usage


class PrintStdOutBase:
    """A :class:`PrintStdOutBase` object is used to print in both the
    stdout and the stdout.txt file, and also to print simple info on
    the current state of the simulation."""

    _tag = "print_stdout"

    @staticmethod
    def _complete_params_with_default(params):
        params.output.periods_print._set_attrib("print_stdout", 1.0)

    def __init__(self, output):
        sim = output.sim
        params = sim.params

        self.output = output
        self.sim = sim
        self.params = params

        try:
            self.c2 = params.c2
            self.f = params.f
            self.nx = params.oper.nx
        except AttributeError:
            pass

        self.period_print = params.output.periods_print.print_stdout

        self.path_file = self.output.path_run + "/stdout.txt"

        if mpi.rank == 0 and self.output._has_to_save:
            if not os.path.exists(self.path_file):
                self.file = open(self.path_file, "w")
            else:
                self.file = open(self.path_file, "r+")
                self.file.seek(0, 2)  # go to the end of the file

    def complete_init_with_state(self):

        self.energy0 = self.output.compute_energy()

        if self.period_print == 0:
            return

        self.energy_temp = self.energy0 + 0.0
        self.t_last_print_info = -self.period_print

        self.print_stdout = self.__call__

    def __call__(self, to_print, end="\n"):
        """Print in stdout and if SAVE in the file stdout.txt"""
        if mpi.rank == 0:
            print(to_print, end=end)
            sys.stdout.flush()
            if self.output._has_to_save:
                self.file.write(to_print + end)
                self.file.flush()
                os.fsync(self.file.fileno())

    def _online_print(self):
        """Print simple info on the current state of the simulation"""
        tsim = self.sim.time_stepping.t
        if (
            tsim + 1e-15
        ) // self.period_print > self.t_last_print_info // self.period_print:
            self._print_info()
            self.t_last_print_info = tsim

    def _print_info(self):
        self.print_stdout(self._make_str_info())
        print_memory_usage("MEMORY_USAGE")

    def _make_str_info(self):
        ts = self.sim.time_stepping
        return (
            f"it = {ts.it:6d} ; t = {ts.t:12.6g} ; deltat  = {ts.deltat:10.5g}\n"
        )

    def _evaluate_duration_left(self):
        """ Computes the remaining time. """
        t_clock = time()
        try:
            delta_clock_time = t_clock - self.t_clock_last
        except AttributeError:
            self.t_clock_last = t_clock
            return

        self.t_clock_last = t_clock
        tsim = self.sim.time_stepping.t

        # variables delta_...: differences between 2 consecutive savings
        delta_equation_time = tsim - self.t_last_print_info

        if delta_equation_time == 0:
            return

        if self.params.time_stepping.USE_T_END:
            remaining_equation_time = self.params.time_stepping.t_end - tsim
        else:
            remaining_equation_time = (
                self.params.time_stepping.it_end - self.sim.time_stepping.it
            ) * self.sim.time_stepping.deltat

        remaining_clock_time = round(
            remaining_equation_time / delta_equation_time * delta_clock_time
        )
        return timedelta(seconds=remaining_clock_time)

    def close(self):
        try:
            self.file.close()
        except AttributeError:
            pass

    def _load_times(self):
        """Load time data from the log file"""
        equation_times = []
        time_steps = []
        remaining_clock_times = []
        times_end = []
        # variables delta_...: differences between 2 consecutive savings
        delta_time_inds = []
        delta_equation_times = []

        time_last = None
        time = None
        it_last = None
        it = None
        equation_time_start = None

        with open(self.output.path_run + "/stdout.txt") as file:
            for line in file:
                if line.startswith("it = "):
                    words = line.split()
                    time_last = time
                    it_last = it
                    it = int(words[2])
                    time = float(words[6])
                    delta_t = float(words[10])

                    if equation_time_start is None:
                        equation_time_start = time

                elif line.startswith("    compute until t ="):
                    t_end = float(line.split()[4])

                elif line.startswith(
                    "              estimated remaining duration"
                ):
                    words = line.split()
                    if "day" in line:
                        day = int(words[4])
                        hour_min_second = words[6]
                    else:
                        day = 0
                        hour_min_second = words[4]

                    nb_hours, nb_minutes, nb_seconds = tuple(
                        int(part) for part in hour_min_second.split(":")
                    )

                    delta = timedelta(
                        days=day,
                        hours=nb_hours,
                        minutes=nb_minutes,
                        seconds=nb_seconds,
                    )

                    remaining_clock_times.append(delta.total_seconds())
                    equation_times.append(time)
                    delta_equation_times.append(time - time_last)
                    delta_time_inds.append(it - it_last)
                    times_end.append(t_end)
                    time_steps.append(delta_t)

        remaining_clock_times = np.array(remaining_clock_times)
        equation_times = np.array(equation_times)
        delta_equation_times = np.array(delta_equation_times)
        times_end = np.array(times_end)
        time_steps = np.array(time_steps)
        delta_time_inds = np.array(delta_time_inds)

        remaining_equation_times = times_end - equation_times
        # see how remaining_clock_time is computed in _evaluate_duration_left
        delta_clock_times = (
            delta_equation_times
            * remaining_clock_times
            / remaining_equation_times
        )
        full_clock_time = delta_clock_times[np.isfinite(delta_clock_times)].sum()

        clock_times_per_timestep = delta_clock_times / delta_time_inds

        loc = locals()
        return {
            key: loc[key]
            for key in (
                "remaining_clock_times",
                "equation_times",
                "delta_equation_times",
                "times_end",
                "time_steps",
                "delta_time_inds",
                "clock_times_per_timestep",
                "equation_time_start",
                "delta_clock_times",
                "full_clock_time",
            )
        }

    def plot_clock_times(self):
        """Plot the estimated full clock time and clock time per time step"""

        results = self._load_times()
        equation_times = results["equation_times"]

        if len(equation_times) == 0:
            print("No time data in the log file. Can't plot anything.")
            return

        fig, axes = plt.subplots(2, 1, sharex=True)

        remaining_clock_times = results["remaining_clock_times"]

        ax = axes[0]
        ax.plot(equation_times, remaining_clock_times)
        ax.set_ylabel("estimated full clock time (s)")

        if remaining_clock_times[-1] > 0.05 * remaining_clock_times.max():
            last_remaining = remaining_clock_times[-1]
            ax.plot(
                equation_times[-1],
                last_remaining,
                "rx",
                label=(
                    f"Last estimation {timedelta(seconds=int(last_remaining))}"
                ),
            )
            ax.legend()

        clock_times_per_timestep = results["clock_times_per_timestep"]
        if clock_times_per_timestep[-1] <= 0.0:
            equation_times = equation_times[:-1]
            clock_times_per_timestep = clock_times_per_timestep[:-1]

        times2 = np.empty(2 * equation_times.size)
        times2[0] = results["equation_time_start"]
        times2[1::2] = equation_times
        times2[2::2] = equation_times[:-1]

        clock_times_per_timestep2 = np.zeros_like(times2)
        clock_times_per_timestep2[::2] = clock_times_per_timestep
        clock_times_per_timestep2[1::2] = clock_times_per_timestep

        ax = axes[1]
        ax.plot(times2, clock_times_per_timestep2)
        ax.set_xlabel("equation time")
        ax.set_ylabel("clock time per time step (s)")
        full_clock_time = results["full_clock_time"]
        ax.set_title(
            f"Full clock time: {timedelta(seconds=int(full_clock_time))}"
        )

        for ax in axes:
            ax.set_ylim(bottom=0)

        fig.suptitle(self.output.summary_simul, fontsize=8)
        fig.tight_layout()



"""Navier-Stokes 2D Boussinesq solvers (:mod:`fluidsim.solvers.ns2d.bouss`)
===========================================================================

Base package for the Navier-Stokes 2D under the Boussinesq approximation
solvers. The main solver defined in :mod:`fluidsim.solvers.ns2d.bouss.solver`
is pseudo-spectral.

The modules for this ns2d.bouss solvers are:

.. autosummary::
   :toctree:

   solver
   state

"""



#!/usr/bin/env python
# coding=utf8

import numpy as np
import glob
import os

import matplotlib.pylab as plt

from solveq2d import solveq2d


# dir_base = 'Approach_runs_2048x2048'
# dir_base = 'Pure_standing_waves_512x512'

dir_base = (
    # '/scratch/augier/'
    "/home/pierre/"
    "Results_for_article_SW1l/"
    "Pure_standing_waves_1024x1024"
)


set_of_dir_results = solveq2d.SetOfDirResults(dir_base=dir_base)

values_c = np.array([10, 1000])

# values_c2 = set_of_dir_results.values_c2
values_c2 = values_c ** 2
values_solver = set_of_dir_results.values_solver
# values_solver = ['SW1l']

tmin = 80


tuple_loop = [
    (c2, name_solver) for c2 in values_c2 for name_solver in values_solver
]
for c2, name_solver in tuple_loop:
    path_dir = set_of_dir_results.one_path_from_values(
        solver=name_solver, c2=c2, FORCING=True
    )

    if path_dir is None:
        continue

    sim = solveq2d.load_state_phys_file(t_approx=2620, name_dir=path_dir)

    sim.output.spect_energy_budg.plot(tmin=tmin, tmax=500, delta_t=0.0)
    sim.output.spectra.plot2D(
        tmin=tmin, tmax=500, delta_t=0.0, coef_compensate=5.0 / 3
    )

    # sim.output.spatial_means.plot()
    sim.output.prob_dens_func.plot(tmin=tmin)

    sim.output.increments.plot(
        tmin=tmin, tmax=None, delta_t=0.0, order=6, yscale="log"
    )

    # sim.output.increments.plot_pdf(tmin=tmin, tmax=160.25,key_var='ux',
    #                                order=4)

    sim.output.increments.plot_Kolmo(tmin=tmin)

    # sim.output.phys_fields.plot(key_field='rot')
    # sim.output.phys_fields.plot(key_field='eta')


solveq2d.show()



"""Module to generate output from Dedalus (:mod:`fluidsim.base.dedalus.output`)
==================================================================================

Provides:

.. autoclass:: OutputDedalus
   :members:
   :private-members:

"""

from ...output.base import OutputBase  # , SpecificOutput


class OutputDedalus(OutputBase):
    @staticmethod
    def _complete_info_solver(info_solver):

        OutputBase._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        base_name_mod = "fluidsim.solvers.ns2d.output"

        classes.PrintStdOut.module_name = base_name_mod + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutNS2D"

        classes.PhysFields.class_name = "PhysFieldsBase2D"

    def compute_energy(self):
        """Compute the spatially averaged energy."""
        vx = self.sim.state.get_var("vx")
        vz = self.sim.state.get_var("vz")
        return 0.5 * (vx ** 2 + vz ** 2).mean()



"""Dedalus solvers (:mod:`fluidsim.base.dedalus`)
=================================================

A reference implementation to use 2D solvers in `Dedalus
<https://dedalus-project.readthedocs.io>`__ via FluidSim.

Provides:

.. autosummary::
   :toctree:

    solver
    state
    operators
    time_stepping
    output

"""



from fluidsim.solvers.ns3d.solver import Simul
import numpy as np

import fluiddyn as fld

params = Simul.create_default_params()

params.short_name_type_run = "test"

n = 64
L = 2 * np.pi
params.oper.nx = n
params.oper.ny = n
params.oper.nz = n
params.oper.Lx = L
params.oper.Ly = L
params.oper.Lz = L
params.oper.type_fft = "fluidfft.fft3d.mpi_with_fftwmpi3d"
# params.oper.type_fft = 'fluidfft.fft3d.with_pyfftw'
# params.oper.type_fft = 'fluidfft.fft3d.with_cufft'

delta_x = params.oper.Lx / params.oper.nx
# params.nu_8 = 2.*10e-1*params.forcing.forcing_rate**(1./3)*delta_x**8
params.nu_8 = 2.0 * 10e-1 * delta_x ** 8

params.time_stepping.USE_T_END = True
params.time_stepping.t_end = 5.0
params.time_stepping.it_end = 2

params.init_fields.type = "dipole"

# params.init_fields.type = 'from_file'
# params.init_fields.from_file.path = '/home/users/bonamy2c/Sim_data/ns3d_test_L=2pix2pix2pi_64x64x64_2016-07-27_16-45-23/state_phys_t=006.006.hd5'

params.forcing.enable = False
# params.forcing.type = 'random'
# 'Proportional'
# params.forcing.type_normalize

params.output.periods_print.print_stdout = 0.00000000001

params.output.periods_save.phys_fields = 1.0
# params.output.periods_save.spectra = 0.5
# params.output.periods_save.spatial_means = 0.05
# params.output.periods_save.spect_energy_budg = 0.5

# params.output.periods_plot.phys_fields = 0.0

params.output.ONLINE_PLOT_OK = True

# params.output.spectra.HAS_TO_PLOT_SAVED = True
# params.output.spatial_means.HAS_TO_PLOT_SAVED = True
# params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True

# params.output.phys_fields.field_to_plot = 'rot'

sim = Simul(params)

# sim.output.phys_fields.plot()
sim.time_stepping.start()
# sim.output.phys_fields.plot()

fld.show()



"""Physical fields output 1d (:mod:`fluidsim.base.output.phys_fields1d`)
========================================================================

Provides:

.. autoclass:: MoviesBasePhysFields1D
   :members:
   :private-members:

.. autoclass:: PhysFieldsBase1D
   :members:
   :private-members:


"""

import numpy as np

from fluiddyn.util import mpi

from .movies import MoviesBase1D

from .phys_fields import PhysFieldsBase


class MoviesBasePhysFields1D(MoviesBase1D):
    def __init__(self, output, phys_fields):
        self.phys_fields = phys_fields
        super().__init__(output)
        self._equation = None


class PhysFieldsBase1D(PhysFieldsBase):
    def _init_movies(self):
        self.movies = MoviesBasePhysFields1D(self.output, self)

    def plot(self, field=None, time=None, numfig=None):
        is_field_ready = False

        key_field = None
        if field is None:
            key_field = self.get_key_field_to_plot(
                forbid_compute=time is not None
            )
        elif isinstance(field, np.ndarray):
            key_field = "given array"
            is_field_ready = True
        elif isinstance(field, str):
            key_field = field

        assert key_field is not None

        if time is None and not is_field_ready:
            # we have to get the field from the state
            time = self.sim.time_stepping.t
            field, _ = self.get_field_to_plot_from_state(key_field)
        else:
            self.set_of_phys_files.update_times()
            # we have to get the field from a file
            if key_field not in self.sim.state.keys_state_phys:
                raise ValueError("key not in state.keys_state_phys")

            field, time = self.get_field_to_plot(key=key_field, time=time)

        if mpi.rank == 0:
            if numfig is None:
                fig, ax = self.output.figure_axe(size_axe=None)
            else:
                fig, ax = self.output.figure_axe(numfig=numfig, size_axe=None)
            xs = self.oper.xs

            ax.plot(xs, field)
            ax.set_xlabel("x")

            self._set_title(ax, key_field, time)

    def _set_title(self, ax, key, time):
        ax.set_title(key + f", $t = {time:.3f}$\n" + self.output.summary_simul)



"""State for the Waves solver (:mod:`fluidsim.solvers.waves2d.state`)
=======================================================================
"""

from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StateWaves(StatePseudoSpectral):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver AD1D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["f_fft", "g_fft"],
                "keys_state_phys": ["f", "g"],
                "keys_computable": [],
                "keys_phys_needed": ["f", "g"],
            }
        )



"""Spatial means regions
========================

.. autoclass:: SpatialMeansRegions
   :members:
   :private-members:

"""

from pathlib import Path
import numbers

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from fluiddyn.util import mpi
from fluidfft.fft3d.operators import vector_product

from fluidsim.base.output.base import SpecificOutput

from . import SimulExtender


class SpatialMeansRegions(SimulExtender, SpecificOutput):
    r"""Specific output for the MILESTONE simulations

    Notes
    -----

    .. |p| mathmacro:: \partial

    .. |vv| mathmacro:: \textbf{v}

    .. |Fh| mathmacro:: \textbf{F}_h

    .. |ez| mathmacro:: \textbf{e}_\textbf{z}

    .. |bnabla| mathmacro:: \boldsymbol{\nabla}

    .. |bomega| mathmacro:: \boldsymbol{\omega}

    This output is used with the solver
    :mod:`fluidsim.solvers.ns3d.strat.solver`, which solves the incompressible
    Navier-Stokes equations (possibly with hyper-viscosity) under the
    Boussinesq approximation with a constant Brunt-Vaisala frequency:

    .. math::

      \p_t \vv + \vv \cdot \bnabla \vv = - \bnabla p + b \ez + D_\vv + \Fh,

      \p_t b   + \vv \cdot \bnabla b = - N^2 v_z  + D_b,

    where :math:`\vv` is the non-divergent velocity (:math:`\bnabla \cdot \vv =
    0`), :math:`p` is the pressure, :math:`b` is the buoyancy, :math:`N` is the
    (constant) Brunt-Vaisala frequency and :math:`D_\vv` and :math:`D_b` are
    dissipative terms.

    The total dynamic pressure :math:`P = p + |\vv|^2/2` can be computed from
    :math:`\vv` and :math:`b` as:

    .. math::

      \bnabla^2 P = \bnabla \cdot (\vv \times \bomega) + \bnabla \cdot \Fh + \p_z b.

    We consider the energy budget

    .. math::

      \p_t |\vv|^2/2 + \bnabla \cdot (\vv P) = \vv \cdot D_\vv + \vv \cdot \Fh + v_z b

      \p_t e_A +  \bnabla \cdot (\vv e_A) = - v_z b + b D_b / N^2,

    with :math:`e_A = b^2/(2N^2)`.

    We take the average of these equations over regions delimited by surfaces
    along the y and z axis. Some terms written as a divergence can be rewritten
    as fluxes through 2 surfaces at :math:`x_{min}` and :math:`x_{max}`:

    .. math::

      \int_V \frac{dV}{V} \bnabla \cdot (\vv P) =
      \frac{S}{V} \left( \langle v_x P \rangle_{S_{min}} - \langle v_x P \rangle_{S_{max}} \right).



    """

    _tag = "spatial_means_regions"
    _module_name = "fluidsim.extend_simul.spatial_means_regions_milestone"

    def __init__(self, output):

        params = output.sim.params

        params_cls = params.output.spatial_means_regions
        self.xmin_given = params_cls.xmin
        self.xmax_given = params_cls.xmax

        if isinstance(self.xmin_given, numbers.Number):
            self.xmin_given = [self.xmin_given]

        if isinstance(self.xmax_given, numbers.Number):
            self.xmax_given = [self.xmax_given]

        self.nb_regions = len(self.xmin_given)

        if len(self.xmax_given) != self.nb_regions:
            raise ValueError("len(self.xmax_given) != len(self.xmin_given)")

        oper = output.sim.oper

        Lx = params.oper.Lx
        if not params.ONLY_COARSE_OPER:
            x_seq = oper.x_seq
        else:
            x_seq = Lx / params.oper.nx * np.arange(params.oper.nx)

        self.info_regions = []

        _, _, ix_seq_start = oper.seq_indices_first_X
        nx_loc = oper.shapeX_loc[2]

        for xmin, xmax in zip(self.xmin_given, self.xmax_given):

            xmin, xmax = Lx * xmin, Lx * xmax

            ixmin = np.argmin(abs(x_seq - xmin))
            xmin = x_seq[ixmin]

            ixmin_loc = ixmin - ix_seq_start
            if ixmin_loc < 0 or ixmin_loc > nx_loc - 1:
                # this limit is not in this process
                ixmin_loc = None

            ixmax = np.argmin(abs(x_seq - xmax))
            xmax = x_seq[ixmax]
            ixmax_loc = ixmax - ix_seq_start
            ixstop_loc = ixmax_loc + 1

            # special case when region == whole numerical domain
            if ixmax == params.oper.nx - 1 and ixmin == 0:
                xmax += x_seq[1]
                ixmax += 1

            ixmax_surf = ixmax % params.oper.nx
            ixmax_surf_loc = ixmax_surf - ix_seq_start
            if ixmax_surf_loc < 0 or ixmax_surf_loc > nx_loc - 1:
                # this limit is not in this process
                ixmax_surf_loc = None

            self.info_regions.append(
                (xmin, xmax, ixmin, ixmax, ixmin_loc, ixmax_surf_loc, ixstop_loc)
            )

        super().__init__(
            output, period_save=params.output.periods_save.spatial_means_regions
        )

        if self.period_save == 0:
            return

        self.masks = []
        self.nb_points = []

        self.nb_points_xmin = []
        self.nb_points_xmax = []

        for info_region in self.info_regions:
            (ixmin_loc, ixmax_surf_loc, ixstop_loc) = info_region[4:]
            mask_loc = np.zeros(shape=oper.shapeX_loc, dtype=np.int8)
            mask_loc[:, :, ixmin_loc:ixstop_loc] = 1
            self.masks.append(mask_loc)

            nb_points = mask_loc.sum()
            if mpi.nb_proc > 1:
                nb_points = mpi.comm.allreduce(nb_points, op=mpi.MPI.SUM)
            self.nb_points.append(nb_points)

            def compute_nb_points_surface(ixsurface_loc):
                if ixsurface_loc is not None:
                    nz_loc = oper.shapeX_loc[0]
                    ny_loc = oper.shapeX_loc[1]
                    nb_points_xsurface_loc = ny_loc * nz_loc
                else:
                    nb_points_xsurface_loc = np.int8(0)
                if mpi.nb_proc > 1:
                    nb_points_xsurface = mpi.comm.allreduce(
                        nb_points_xsurface_loc, op=mpi.MPI.SUM
                    )
                else:
                    nb_points_xsurface = nb_points_xsurface_loc

                return nb_points_xsurface

            self.nb_points_xmin.append(compute_nb_points_surface(ixmin_loc))
            self.nb_points_xmax.append(compute_nb_points_surface(ixmax_surf_loc))

        self._save_one_time()

    def _init_path_files(self):
        self.path_dir = Path(self.output.path_run) / self._tag
        self.paths = [
            self.path_dir / f"data{iregion}.csv"
            for iregion in range(self.nb_regions)
        ]

    def _init_files(self, arrays_1st_time=None):
        if mpi.rank == 0:
            self.path_dir.mkdir(exist_ok=True)
            for path, info_region in zip(self.paths, self.info_regions):
                xmin, xmax = info_region[:2]
                if not path.exists():
                    with open(path, "w") as file:
                        file.write(
                            f"# xmin = {xmin} ; xmax = {xmax}\n"
                            "time,EK,EKz,EA,epsK,epsA,conv_K2A,PK,PA,"
                            "flux_Pnl_xmin,flux_Pnl_xmax,flux_v2_xmin,flux_v2_xmax,"
                            "flux_P_dz_b_xmin,flux_P_dz_b_xmax,"
                            "flux_Pforcing_xmin,flux_Pforcing_xmax,"
                            "flux_A_xmin,flux_A_xmax\n"
                        )
                else:
                    with open(path, "r") as file:
                        words = file.readline().split()
                        xmin_file = float(words[3])
                        xmax_file = float(words[7])
                        if xmin_file != xmin or xmax_file != xmax:
                            raise ValueError(
                                "xmin_file != xmin or xmax_file != xmax\n"
                                f"{xmin_file} != {xmin} or {xmax_file} != {xmax}"
                            )

    @classmethod
    def complete_params_with_default(cls, params):
        params.output.periods_save._set_attrib(cls._tag, 0)
        params.output._set_child(
            cls._tag, attribs={"xmin": [0.1, 0.4, 0.7], "xmax": [0.3, 0.6, 0.9]}
        )

    @classmethod
    def get_modif_info_solver(cls):
        """Create a function to modify ``info_solver``.

        Note that this function is called when the object ``info_solver`` has
        not yet been created (and cannot yet be modified)! This is why one
        needs to create a function that will be called later to modify
        ``info_solver``.

        """

        def modif_info_solver(info_solver):
            info_solver.classes.Output.classes._set_child(
                "SpatialMeansRegions",
                attribs={
                    "module_name": cls._module_name,
                    "class_name": cls.__name__,
                },
            )

        return modif_info_solver

    def _online_save(self):
        if self._has_to_online_save():
            self._save_one_time()

    def _compute_means_regions(self, field):
        results = []
        for nb_points, mask in zip(self.nb_points, self.masks):
            result = (mask * field).sum()
            if mpi.nb_proc > 1:
                result = mpi.comm.allreduce(result, op=mpi.MPI.SUM)
            results.append(result / nb_points)
        return np.array(results)

    def _compute_fluxes_regions(self, field, vx):
        """Compute for each region the fluxes over xmin and xmax surface"""
        fluxes = []
        for index_region, info_region in enumerate(self.info_regions):
            xmin, xmax = info_region[:2]
            length_region = xmax - xmin
            ixmin_loc, ixmax_surf_loc = info_region[4:-1]

            def compute_flux(nb_points_surfaces, ixsurface_loc):
                nb_points_surface = nb_points_surfaces[index_region]
                if ixsurface_loc is not None:
                    field_xsurface = field[:, :, ixsurface_loc]
                    vx_xsurface = vx[:, :, ixsurface_loc]
                    sum_xsurface = (vx_xsurface * field_xsurface).sum()
                else:
                    sum_xsurface = 0.0
                if mpi.nb_proc > 1:
                    sum_xsurface = mpi.comm.allreduce(
                        sum_xsurface, op=mpi.MPI.SUM
                    )
                return sum_xsurface / nb_points_surface / length_region

            flux_xmin = compute_flux(self.nb_points_xmin, ixmin_loc)
            flux_xmax = -compute_flux(self.nb_points_xmax, ixmax_surf_loc)
            fluxes.append((flux_xmin, flux_xmax))
        return np.array(fluxes)

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim

        state = self.sim.state
        oper = self.sim.oper

        ifft = oper.ifft
        fft = oper.fft
        ifft_as_arg_destroy = oper.ifft_as_arg_destroy

        get_var = state.state_phys.get_var
        b = get_var("b")
        vx = get_var("vx")
        vy = get_var("vy")
        vz = get_var("vz")

        N2b2 = 0.5 / self.sim.params.N ** 2 * b * b
        EAs = self._compute_means_regions(N2b2)

        vh2 = 0.5 * (vx * vx + vy * vy)
        vz2 = 0.5 * vz * vz
        v2_over_2 = vh2 + vz2

        EKhs = self._compute_means_regions(vh2)
        EKzs = self._compute_means_regions(vz2)
        del vh2, vz2

        EKs = EKhs + EKzs

        conv_K2A = -self._compute_means_regions(b * vz)

        get_var = state.state_spect.get_var
        b_fft = get_var("b_fft")
        vx_fft = get_var("vx_fft")
        vy_fft = get_var("vy_fft")
        vz_fft = get_var("vz_fft")

        f_d, _ = self.sim.compute_freq_diss()

        f_d_vx = ifft(f_d * vx_fft)
        f_d_vy = ifft(f_d * vy_fft)
        f_d_vz = ifft(f_d * vz_fft)
        epsKs = self._compute_means_regions(
            vx * f_d_vx + vy * f_d_vy + vz * f_d_vz
        )
        del f_d_vx, f_d_vy, f_d_vz

        epsAs = [
            epsA / self.sim.params.N ** 2
            for epsA in self._compute_means_regions(b * ifft(f_d * b_fft))
        ]

        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            forcing_fft = self.sim.forcing.get_forcing()

            fx_fft = forcing_fft.get_var("vx_fft")
            fy_fft = forcing_fft.get_var("vy_fft")

            assert np.allclose(forcing_fft.get_var("vz_fft"), 0)

            fx = ifft(fx_fft)
            fy = ifft(fy_fft)

            PKs = self._compute_means_regions(
                vx * fx + vy * fy + (abs(fx) ** 2 + abs(fy) ** 2) * deltat / 2
            )

        else:
            PKs = np.zeros(3)

        # Compute spatial fluxes
        # Need to compute the pressure P = v^2/2 + p

        omegax_fft, omegay_fft, omegaz_fft = oper.rotfft_from_vecfft(
            vx_fft, vy_fft, vz_fft
        )

        ifft_as_arg_destroy = oper.ifft_as_arg_destroy

        omegax = state.fields_tmp[3]
        omegay = state.fields_tmp[4]
        omegaz = state.fields_tmp[5]

        ifft_as_arg_destroy(omegax_fft, omegax)
        ifft_as_arg_destroy(omegay_fft, omegay)
        ifft_as_arg_destroy(omegaz_fft, omegaz)

        del omegax_fft, omegay_fft, omegaz_fft

        f_nl_x, f_nl_y, f_nl_z = vector_product(
            vx, vy, vz, omegax, omegay, omegaz
        )

        del omegax, omegay, omegaz

        f_nl_x_fft = fft(f_nl_x)
        f_nl_y_fft = fft(f_nl_y)
        f_nl_z_fft = fft(f_nl_z)

        f_nl_x_fft *= 1j * oper.Kx
        dx_f_nl_x_fft = f_nl_x_fft
        del f_nl_x_fft

        f_nl_y_fft *= 1j * oper.Ky
        dy_f_nl_y_fft = f_nl_y_fft
        del f_nl_y_fft

        f_nl_z_fft *= 1j * oper.Kz
        dz_f_nl_z_fft = f_nl_z_fft
        del f_nl_z_fft

        P_nl_fft = -(dx_f_nl_x_fft + dy_f_nl_y_fft + dz_f_nl_z_fft) / oper.K2_not0
        del dx_f_nl_x_fft, dy_f_nl_y_fft, dz_f_nl_z_fft
        P_nl = ifft(P_nl_fft)
        del P_nl_fft

        if self.sim.params.forcing.enable:
            fz_fft = forcing_fft.get_var("vz_fft")
            P_forcing = -ifft(
                oper.divfft_from_vecfft(fx_fft, fy_fft, fz_fft) / oper.K2_not0
            )

        P_dz_b = -ifft(1j * oper.Kz * b_fft / oper.K2_not0)

        fluxes_P_nl = self._compute_fluxes_regions(P_nl, vx)
        fluxes_v2 = self._compute_fluxes_regions(v2_over_2, vx)
        fluxes_P_forcing = self._compute_fluxes_regions(P_forcing, vx)
        fluxes_P_dz_b = self._compute_fluxes_regions(P_dz_b, vx)
        fluxes_A = self._compute_fluxes_regions(N2b2, vx)

        if mpi.rank > 0:
            return

        for i, path in enumerate(self.paths):
            flux_Pnl_xmin, flux_Pnl_xmax = fluxes_P_nl[i]
            flux_v2_xmin, flux_v2_xmax = fluxes_v2[i]
            flux_Pforcing_xmin, flux_Pforcing_xmax = fluxes_P_forcing[i]
            flux_P_dz_b_xmin, flux_P_dz_b_xmax = fluxes_P_dz_b[i]
            flux_A_xmin, flux_A_xmax = fluxes_A[i]

            with open(path, "a") as file:
                file.write(
                    f"{tsim},{EKs[i]},{EKzs[i]},{EAs[i]},"
                    f"{epsKs[i]},{epsAs[i]},{conv_K2A[i]},{PKs[i]},0.0,"
                    f"{flux_Pnl_xmin},{flux_Pnl_xmax},{flux_v2_xmin},{flux_v2_xmax},"
                    f"{flux_P_dz_b_xmin},{flux_P_dz_b_xmax},"
                    f"{flux_Pforcing_xmin},{flux_Pforcing_xmax},"
                    f"{flux_A_xmin},{flux_A_xmax}\n"
                )

    def load(self, iregion=0):
        df = pd.read_csv(self.paths[iregion], skiprows=1)
        return df

    def print_keys(self):
        """Print the keys associated with the computed quantities"""
        df = pd.read_csv(self.paths[0], skiprows=1, nrows=1)
        print(df.columns)

    def plot(self, keys="EK", iregion=0):
        """Plot some quantities for a given region"""
        df = self.load(iregion)
        ax = df.plot(x="time", y=keys)
        xmin, xmax = self.info_regions[iregion][:2]
        ax.set_title(self.output.summary_simul + f"\nxmin={xmin}, xmax={xmax}")
        return ax

    def plot_budget(
        self, iregion=0, decompose_fluxes=False, plot_conversion=False
    ):
        """Plot the energy budget for a given region"""
        df = self.load(iregion)
        times = df["time"]

        E_tot = df["EK"] + df["EA"]
        dt_E_tot = np.gradient(E_tot, times)

        P_tot = df["PK"] + df["PA"]
        eps = df["epsK"] + df["epsA"]

        fig, ax = plt.subplots()

        ax.plot(times, dt_E_tot, "k--", label="$d_t E$", linewidth=2)

        ax.plot(times, P_tot, label="Forcing")
        ax.plot(times, -eps, label="Viscosity")

        kinds = "flux_Pnl flux_Pforcing flux_A flux_P_dz_b".split()

        flux_tot = np.zeros_like(times)

        for kind in kinds:
            flux_kind = df[kind + "_xmin"] + df[kind + "_xmax"]
            flux_tot += flux_kind.values

        ax.plot(times, flux_tot, label="Surface fluxes")
        ax.plot(times, P_tot - eps + flux_tot, label="All terms", linewidth=2)

        if decompose_fluxes:
            for kind in kinds:
                flux_kind = df[kind + "_xmin"] + df[kind + "_xmax"]
                ax.plot(times, flux_kind, ":", label=kind)

        if plot_conversion:
            ax.plot(times, df["conv_K2A"], "-.", label=r"$C_{K\rightarrow A}$")

        xmin, xmax = self.info_regions[iregion][:2]
        ax.set_title(
            self.output.summary_simul + f"\nxmin={xmin:.3f}, xmax={xmax:.3f}"
        )
        ax.set_xlabel("time")

        fig.legend()
        fig.tight_layout()

        return ax



import numpy as np
import matplotlib.pylab as plt

from solveq2d import solveq2d

name_dir = (
    "SE2D_SW1lwaves_forcingw_L=50.x50._256x256_c=50_f=0_2013-06-14_09-48-06"
)
sim = solveq2d.create_sim_plot_from_dir(name_dir)

nx = sim.param.nx
c2 = sim.param.c2
f = sim.param.f
name_solver = sim.output.name_solver

dict_results = sim.output.spatial_means.load()


def derivate1D(t, func):
    dt_func = np.empty(func.shape, dtype=func.dtype)
    dt_func[1:-1] = (func[2:] - func[0:-2]) / (t[2:] - t[0:-2])
    dt_func[0] = (func[1] - func[0]) / (t[1] - t[0])
    dt_func[-1] = (func[-1] - func[-2]) / (t[-1] - t[-2])
    return dt_func


t = dict_results["t"]

E = dict_results["E"]
CPE = dict_results["CPE"]

EK = dict_results["EK"]
EA = dict_results["EA"]
EKr = dict_results["EKr"]

epsK = dict_results["epsK"]
epsK_hypo = dict_results["epsK_hypo"]
epsK_tot = dict_results["epsK_tot"]

epsA = dict_results["epsA"]
epsA_hypo = dict_results["epsA_hypo"]
epsA_tot = dict_results["epsA_tot"]

epsE = epsK + epsA
epsE_hypo = epsK_hypo + epsA_hypo
epsE_tot = epsK_tot + epsA_tot

epsCPE = dict_results["epsCPE"]
epsCPE_hypo = dict_results["epsCPE_hypo"]
epsCPE_tot = dict_results["epsCPE_tot"]

PK_tot = dict_results["PK_tot"]
PA_tot = dict_results["PA_tot"]
P_tot = PK_tot + PA_tot


width_axe = 0.85
height_axe = 0.37
x_left_axe = 0.12
z_bottom_axe = 0.56

size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
fig, ax1 = sim.output.figure_axe(size_axe=size_axe)
ax1.set_xlabel("$t$")
ax1.set_ylabel("$2E(t)/c^2$")
title = (
    "mean energy, solver "
    + sim.output.name_solver
    + f", nh = {nx:5d}"
    + f", c2 = {c2:.4g}, f = {f:.4g}"
)

ax1.set_title(title)
ax1.hold(True)
norm = c2 / 2
ax1.plot(t, E / norm, "k", linewidth=2)
ax1.plot(t, EK / norm, "r", linewidth=1)
ax1.plot(t, EA / norm, "b", linewidth=1)
ax1.plot(t, EKr / norm, "r--", linewidth=1)
ax1.plot(t, (EK - EKr) / norm, "r:", linewidth=1)

z_bottom_axe = 0.07
size_axe[1] = z_bottom_axe
ax2 = fig.add_axes(size_axe)
ax2.set_xlabel("t")
ax2.set_ylabel("Charney PE(t)")
title = "mean Charney PE(t)"
ax2.set_title(title)
ax2.hold(True)
ax2.plot(t, CPE, "k", linewidth=2)


z_bottom_axe = 0.56
size_axe[1] = z_bottom_axe
fig, ax1 = sim.output.figure_axe(size_axe=size_axe)
ax1.set_xlabel("t")
ax1.set_ylabel(r"$P_E(t)$, $\epsilon(t)$")
title = (
    "forcing and dissipation, solver "
    + sim.output.name_solver
    + f", nh = {nx:5d}"
    + f", c2 = {c2:.4g}, f = {f:.4g}"
)
ax1.set_title(title)
ax1.hold(True)
(l_P_tot,) = ax1.plot(t, P_tot, "c", linewidth=2)
(l_epsE,) = ax1.plot(t, epsE, "k--", linewidth=2)
(l_epsE_hypo,) = ax1.plot(t, epsE_hypo, "g", linewidth=2)
(l_epsE_tot,) = ax1.plot(t, epsE_tot, "k", linewidth=2)
l_P_tot.set_label("$P_{tot}$")
l_epsE.set_label(r"$\epsilon$")
l_epsE_hypo.set_label(r"$\epsilon_{hypo}$")
l_epsE_tot.set_label(r"$\epsilon_{tot}$")

ax1.legend(loc=2)

z_bottom_axe = 0.07
size_axe[1] = z_bottom_axe
ax2 = fig.add_axes(size_axe)
ax2.set_xlabel("t")
ax2.set_ylabel(r"$\epsilon$ Charney PE(t)")
title = "dissipation Charney PE"
ax2.set_title(title)
ax2.hold(True)
ax2.plot(t, epsCPE, "k--", linewidth=2)
ax2.plot(t, epsCPE_hypo, "g", linewidth=2)
ax2.plot(t, epsCPE_tot, "r", linewidth=2)

if dict_results.has_key("epsKsuppl"):
    epsKsuppl = dict_results["epsKsuppl"]
    epsKsuppl_hypo = dict_results["epsKsuppl_hypo"]

    ax1.plot(t, epsKsuppl, "r--", linewidth=1)
    ax1.plot(t, epsKsuppl_hypo, "g--", linewidth=1)

if "Conv" in dict_results.keys():
    print("Conv is saved")
    Conv = dict_results["Conv"]
    c2eta1d = dict_results["c2eta1d"]
    c2eta2d = dict_results["c2eta2d"]
    c2eta3d = dict_results["c2eta3d"]

    Conv2 = c2eta1d + c2eta2d / 2

    eps_mean = epsE.mean()
    E_mean = E.mean()

    # fig, ax1 = sim.output.figure_axe()

    # title = ('Conv, solver '+sim.output.name_solver+
    #      ', nh = {0:5d}'.format(nx)+
    #      ', c2 = {0:.4g}, f = {1:.4g}'.format(c2, f)
    #          )
    # ax1.set_title(title)
    # ax1.set_xlabel('t')

    # ax1.plot(t, Conv/eps_mean, 'k', linewidth=2)
    # ax1.plot(t, ConvA/eps_mean, 'k--', linewidth=2)

    print("E_mean  =", E_mean)
    print("eps_mean  =", eps_mean)
    print("Conv.mean()/eps_mean  =", Conv.mean() / eps_mean)
    print("Conv2.mean()/eps_mean  =", Conv2.mean() / eps_mean)
    print("c2eta1d.mean()/eps_mean  =", c2eta1d.mean() / eps_mean)
    print("c2eta2d.mean()/eps_mean  =", c2eta2d.mean() / eps_mean)
    print("c2eta3d.mean()/eps_mean  =", c2eta3d.mean() / eps_mean)

    # norm = E_mean**2

    # ax1.plot(t, cetaConv/norm, 'r', linewidth=2)
    # ax1.plot(t, cetaConvA/norm, 'r--', linewidth=2)

    # print('cetaConv.mean()/norm  =', cetaConv.mean()/norm)
    # print('cetaConvA.mean()/norm =', cetaConvA.mean()/norm)

solveq2d.show()



import os

import numpy as np
import h5py

from fluiddyn.util import mpi

from .base import SpecificOutput


class SpectraMultiDim(SpecificOutput):
    """Saving the multidimensional spectra."""

    _tag = "spectra_multidim"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spectra_multidim"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def __init__(self, output):
        self.output = output

        params = output.sim.params
        self.nx = int(params.oper.nx)

        if not params.output.HAS_TO_SAVE:
            params.output.periods_save.spectra_multidim = False

        super().__init__(
            output,
            period_save=params.output.periods_save.spectra_multidim,
            has_to_plot_saved=params.output.spectra_multidim.HAS_TO_PLOT_SAVED,
        )

    def _init_path_files(self):
        path_run = self.output.path_run
        self.path_file = path_run + "/spectra_multidim.h5"

    def _init_files(self, arrays_1st_time=None):
        dict_spectra = self.compute()
        if mpi.rank == 0:
            if not os.path.exists(self.path_file):
                arrays_1st_time = {
                    "kxE": self.sim.oper.kxE,
                    "kyE": self.sim.oper.kyE,
                }
                self._create_file_from_dict_arrays(
                    self.path_file, dict_spectra, arrays_1st_time
                )
                self.nb_saved_times = 1
            else:
                with h5py.File(self.path_file, "r") as file:
                    dset_times = file["times"]
                    self.nb_saved_times = dset_times.shape[0] + 1
                # save the spectra in the file spectra_multidim.h5
                self._add_dict_arrays_to_file(self.path_file, dict_spectra)

        self.t_last_save = self.sim.time_stepping.t

    def _online_save(self):
        """Save the values at one time. """
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_save >= self.period_save:
            self.t_last_save = tsim
            dict_spectra = self.compute()
            if mpi.rank == 0:
                # save the spectra in the file spectra_multidim.h5
                self._add_dict_arrays_to_file(self.path_file, dict_spectra)
                self.nb_saved_times += 1
                if self.has_to_plot:
                    self._online_plot_saving(dict_spectra)

                    if tsim - self.t_last_show >= self.period_show:
                        self.t_last_show = tsim
                        self.axe.get_figure().canvas.draw()

    def compute(self):
        """compute the values at one time."""
        if mpi.rank == 0:
            dict_results = {}
            return dict_results

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, axe = self.output.figure_axe(numfig=1_000_000)
            self.axe = axe
            axe.set_xlabel("$k_x$")
            axe.set_ylabel("$k_y$")
            axe.set_title(
                "Multidimensional spectra\n" + self.output.summary_simul
            )

    def _online_plot_saving(self, arg):
        pass

    def load_mean(self, tmin=None, tmax=None):
        """Loads time averaged data between tmin and tmax."""

        results = {}

        # Load data
        with h5py.File(self.path_file, "r") as file:
            for key in file.keys():
                if not key.startswith("info"):
                    results[key] = file[key][...]

        # Time average spectra
        times = results["times"]
        nt = len(times)
        if tmin is None:
            imin_plot = 0
        else:
            imin_plot = np.argmin(abs(times - tmin))

        if tmax is None:
            imax_plot = nt - 1
        else:
            imax_plot = np.argmin(abs(times - tmax))

        tmin = times[imin_plot]
        tmax = times[imax_plot]

        print(
            "compute mean of multidimensional spectra\n"
            f"tmin = {tmin:8.6g} ; tmax = {tmax:8.6g}\n"
            f"imin = {imin_plot:8d} ; imax = {imax_plot:8d}"
        )

        for key in results.keys():
            if key.startswith("spectr"):
                spect = results[key]
                spect_averaged = spect[imin_plot : imax_plot + 1].mean(0)
                results[key] = spect_averaged

        return results

    def plot(self):
        pass



"""Predator-prey solver (:mod:`fluidsim.solvers.models0d.predaprey.solver`)
===========================================================================

This module provides classes to solve the Lotka-Volterra equations.

.. autoclass:: InfoSolverPredaPrey
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:

.. autoclass:: StatePredaPrey
   :members:
   :private-members:

"""

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.base.solvers.base import InfoSolverBase, SimulBase
from fluidsim.base.state import StateBase


class StatePredaPrey(StateBase):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""
        info_solver.classes.State._set_attribs(
            {
                "keys_state_phys": ["X", "Y"],
                "keys_computable": [],
                "keys_phys_needed": ["X", "Y"],
                "keys_linear_eigenmodes": ["X", "Y"],
            }
        )


class InfoSolverPredaPrey(InfoSolverBase):
    """Contain the information on the solver predaprey."""

    def _init_root(self):
        super()._init_root()

        package = "fluidsim.solvers.models0d.predaprey"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "predaprey"

        classes = self.classes

        classes.State.module_name = package + ".solver"
        classes.State.class_name = "StatePredaPrey"

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "Output"


class Simul(SimulBase):
    """Solve the Lotka-Volterra equations."""

    InfoSolver = InfoSolverPredaPrey

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulBase._complete_params_with_default(params)
        attribs = {"A": 1.0, "B": 1.0, "C": 1.0, "D": 0.5}
        params._set_attribs(attribs)

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        p = self.params
        self.Xs = p.C / p.D
        self.Ys = p.A / p.B

    def tendencies_nonlin(self, state=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state.  If `state is not None`, the variables
            are computed from it, otherwise, they are taken from the global
            state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the variables.

        Notes
        -----

        The Lotka-Volterra equations can be written

        .. math::
           \dot X = AX - B XY,
           \dot Y = -CY + D XY.


        """
        p = self.params

        if state is None:
            state = self.state.state_phys

        X = state.get_var("X")
        Y = state.get_var("Y")

        if old is None:
            tendencies = SetOfVariables(like=self.state.state_phys)
        else:
            tendencies = old
        tendencies.set_var("X", p.A * X - p.B * X * Y)
        tendencies.set_var("Y", -p.C * Y + p.D * X * Y)

        if self.params.forcing.enable:
            tendencies += self.forcing.get_forcing()

        return tendencies


if __name__ == "__main__":
    import fluiddyn as fld

    params = Simul.create_default_params()

    params.time_stepping.deltat0 = 0.1
    params.time_stepping.t_end = 20

    params.output.periods_print.print_stdout = 0.01

    sim = Simul(params)

    sim.state.state_phys.set_var("X", 2)
    sim.state.state_phys.set_var("Y", 1.1)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    fld.show()



from warnings import warn
import numpy as np


def load_txt(path_file, dict_results={}):
    warn(
        "Kept for legacy reasons to read old simulations, could be removed in the future",
        DeprecationWarning,
    )

    with open(path_file) as file_means:
        lines = file_means.readlines()

    lines_t = []
    lines_E = []
    lines_EK = []
    lines_epsK = []
    lines_epsA = []
    lines_epsCPE = []

    lines_epsK = []

    lines_PK = []
    lines_PA = []
    lines_etaskew = []
    lines_rotskew = []
    lines_Conv = []
    # Supplimentary dissipation
    lines_epsKsuppl = []

    for il, line in enumerate(lines):
        if line[0:6] == "time =":
            lines_t.append(line)
        if line[0:8] == "E      =":
            lines_E.append(line)
        if line[0:8] == "EK     =":
            lines_EK.append(line)
        if line[0:8] == "epsK   =":
            lines_epsK.append(line)
        if line[0:8] == "epsA   =":
            lines_epsA.append(line)
        if line[0:8] == "epsCPE =":
            lines_epsCPE.append(line)
        if line[0:8] == "PK1    =":
            lines_PK.append(line)
        if line[0:8] == "PA1    =":
            lines_PA.append(line)
        if line.startswith("eta skew ="):
            lines_etaskew.append(line)
        if line.startswith("rot skew ="):
            lines_rotskew.append(line)
        if line.startswith("Conv ="):
            lines_Conv.append(line)

        if line.startswith("epsKsup="):
            lines_epsKsuppl.append(line)

    nt = len(lines_t)
    # NOTE: unnecessary removal of last record
    #  if nt > 1:
    #      nt -= 1

    t = np.empty(nt)

    E = np.empty(nt)
    CPE = np.empty(nt)
    EK = np.empty(nt)
    EA = np.empty(nt)
    EKr = np.empty(nt)
    epsK = np.empty(nt)
    epsK_hypo = np.empty(nt)
    epsK_tot = np.empty(nt)
    epsA = np.empty(nt)
    epsA_hypo = np.empty(nt)
    epsA_tot = np.empty(nt)
    epsCPE = np.empty(nt)
    epsCPE_hypo = np.empty(nt)
    epsCPE_tot = np.empty(nt)

    if len(lines_PK) == len(lines_t):
        PK1 = np.empty(nt)
        PK2 = np.empty(nt)
        PK_tot = np.empty(nt)
        PA1 = np.empty(nt)
        PA2 = np.empty(nt)
        PA_tot = np.empty(nt)

    if len(lines_rotskew) == len(lines_t):
        skew_eta = np.empty(nt)
        kurt_eta = np.empty(nt)
        skew_rot = np.empty(nt)
        kurt_rot = np.empty(nt)

    if len(lines_Conv) == len(lines_t):
        Conv = np.empty(nt)
        c2eta1d = np.empty(nt)
        c2eta2d = np.empty(nt)
        c2eta3d = np.empty(nt)

    if len(lines_epsKsuppl) == len(lines_t):
        epsKsuppl = np.empty(nt)
        epsKsuppl_hypo = np.empty(nt)

    for il in range(nt):
        line = lines_t[il]
        words = line.split()
        t[il] = float(words[2])

        line = lines_E[il]
        words = line.split()
        E[il] = float(words[2])
        CPE[il] = float(words[6])

        line = lines_EK[il]
        words = line.split()
        EK[il] = float(words[2])
        EA[il] = float(words[6])
        EKr[il] = float(words[10])

        line = lines_epsK[il]
        words = line.split()
        epsK[il] = float(words[2])
        epsK_hypo[il] = float(words[6])
        epsK_tot[il] = float(words[10])

        line = lines_epsA[il]
        words = line.split()
        epsA[il] = float(words[2])
        epsA_hypo[il] = float(words[6])
        epsA_tot[il] = float(words[10])

        line = lines_epsCPE[il]
        words = line.split()
        epsCPE[il] = float(words[2])
        epsCPE_hypo[il] = float(words[6])
        epsCPE_tot[il] = float(words[10])

        if len(lines_PK) == len(lines_t):
            line = lines_PK[il]
            words = line.split()
            PK1[il] = float(words[2])
            PK2[il] = float(words[6])
            PK_tot[il] = float(words[10])

            line = lines_PA[il]
            words = line.split()
            PA1[il] = float(words[2])
            PA2[il] = float(words[6])
            PA_tot[il] = float(words[10])

        if len(lines_rotskew) == len(lines_t):
            line = lines_etaskew[il]
            words = line.split()
            skew_eta[il] = float(words[3])
            kurt_eta[il] = float(words[7])

            line = lines_rotskew[il]
            words = line.split()
            skew_rot[il] = float(words[3])
            kurt_rot[il] = float(words[7])

        if len(lines_Conv) == len(lines_t):
            line = lines_Conv[il]
            words = line.split()
            Conv[il] = float(words[2])
            c2eta1d[il] = float(words[6])
            c2eta2d[il] = float(words[10])
            c2eta3d[il] = float(words[14])

        if len(lines_epsKsuppl) == len(lines_t):
            line = lines_epsKsuppl[il]
            words = line.split()
            epsKsuppl[il] = float(words[1])
            epsKsuppl_hypo[il] = float(words[5])

    dict_results["t"] = t
    dict_results["E"] = E
    dict_results["CPE"] = CPE

    dict_results["EK"] = EK
    dict_results["EA"] = EA
    dict_results["EKr"] = EKr

    dict_results["epsK"] = epsK
    dict_results["epsK_hypo"] = epsK_hypo
    dict_results["epsK_tot"] = epsK_tot

    dict_results["epsA"] = epsA
    dict_results["epsA_hypo"] = epsA_hypo
    dict_results["epsA_tot"] = epsA_tot

    dict_results["epsCPE"] = epsCPE
    dict_results["epsCPE_hypo"] = epsCPE_hypo
    dict_results["epsCPE_tot"] = epsCPE_tot

    if len(lines_PK) == len(lines_t):
        dict_results["PK1"] = PK1
        dict_results["PK2"] = PK2
        dict_results["PK_tot"] = PK_tot
        dict_results["PA1"] = PA1
        dict_results["PA2"] = PA2
        dict_results["PA_tot"] = PA_tot

    if len(lines_rotskew) == len(lines_t):
        dict_results["skew_eta"] = skew_eta
        dict_results["kurt_eta"] = kurt_eta
        dict_results["skew_rot"] = skew_rot
        dict_results["kurt_rot"] = kurt_rot

    if len(lines_Conv) == len(lines_t):
        dict_results["Conv"] = Conv
        dict_results["c2eta1d"] = c2eta1d
        dict_results["c2eta2d"] = c2eta2d
        dict_results["c2eta3d"] = c2eta3d

    if len(lines_epsKsuppl) == len(lines_t):
        dict_results["epsKsuppl"] = epsKsuppl
        dict_results["epsKsuppl_hypo"] = epsKsuppl_hypo

    return dict_results



#!/usr/bin/env python
# coding=utf8

import numpy as np
import glob
import os

import matplotlib.pylab as plt

from solveq2d import solveq2d

LOAD = True
# LOAD = False

dir_base = "Approach_runs_1024x1024"
# dir_base = 'Waves_statio_256x256'

set_of_dir_results = solveq2d.SetOfDirResults(dir_base=dir_base)

values_c2 = set_of_dir_results.values_c2
# values_c2 = [600]
values_solver = set_of_dir_results.values_solver
values_solver = ["SW1l"]

tstatio = 250


if LOAD:

    tuple_loop = [
        (c2, name_solver) for c2 in values_c2 for name_solver in values_solver
    ]

    nb_runs = 0
    for c2, name_solver in tuple_loop:
        nb_runs += 1

    arr_c2 = np.empty([nb_runs])
    Froude_numbers = np.empty([nb_runs])
    EK = np.empty([nb_runs])
    EA = np.empty([nb_runs])
    EKr = np.empty([nb_runs])
    epsK = np.empty([nb_runs])
    epsA = np.empty([nb_runs])
    epsK_tot = np.empty([nb_runs])
    epsA_tot = np.empty([nb_runs])

    irun = -1

    for c2, name_solver in tuple_loop:
        path_dir = set_of_dir_results.one_path_from_values(
            solver=name_solver, c2=c2
        )

        irun += 1
        sim = solveq2d.create_sim_plot_from_dir(name_dir=path_dir)

        (
            dict_time_means,
            dict_results,
        ) = sim.output.spatial_means.compute_time_means(tstatio)

        arr_c2[irun] = sim.param["c2"]
        EK[irun] = dict_time_means["EK"]

        Froude_numbers[irun] = 2 * EK[irun] / sim.param["c2"]

        EA[irun] = dict_time_means["EA"]
        EKr[irun] = dict_time_means["EKr"]
        epsK[irun] = dict_time_means["epsK"]
        epsA[irun] = dict_time_means["epsA"]
        epsK_tot[irun] = dict_time_means["epsK_tot"]
        epsA_tot[irun] = dict_time_means["epsA_tot"]

    eps = epsK + epsA
    eps_tot = epsK_tot + epsA_tot


# width_axe = 0.85
# height_axe = 0.37
# x_left_axe = 0.12
# z_bottom_axe = 0.56

# size_axe = [x_left_axe, z_bottom_axe,
#             width_axe, height_axe]
# fig, ax1 = sim.output.figure_axe(size_axe=size_axe)

fig, ax1 = sim.output.figure_axe()

ax1.set_xlabel("$2E/c^2$")
ax1.set_ylabel(r"$\epsilon / P$")
ax1.hold(True)

coef_norm = Froude_numbers ** -1.54
ax1.loglog(Froude_numbers, Froude_numbers ** 1.5 * coef_norm, "k--", linewidth=2)

# coef_norm = Froude_numbers**-3
# ax1.loglog(Froude_numbers, Froude_numbers**3*coef_norm, 'k--', linewidth=2)

ax1.loglog(Froude_numbers, eps / eps_tot * coef_norm, "kx", linewidth=2)


plt.show()



import argparse
import fluidsim as fls

parser = argparse.ArgumentParser(
    description="Resume a Fluidsim simulation from a path"
)
parser.add_argument("path", help="path to an incomplete simulation directory")

args = parser.parse_args()
sim = fls.load_state_phys_file(args.path, modif_save_params=False)
sim.time_stepping.start()



"""Variable container (:mod:`fluidsim.base.setofvariables`)
===========================================================

Provides:

.. autoclass:: SetOfVariables
   :members:
   :private-members:

"""

from fluiddyn.calcul.setofvariables import SetOfVariables



"""Output (:mod:`fluidsim.solvers.ns2d.output`)
===============================================

Provides the modules:

.. autosummary::
   :toctree:

   print_stdout
   spatial_means
   spectra
   spect_energy_budget

and the main output class for the ns2d solver:

.. autoclass:: Output
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.output import OutputBasePseudoSpectral


class Output(OutputBasePseudoSpectral):
    """Output for ns2d solver."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""

        OutputBasePseudoSpectral._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        base_name_mod = "fluidsim.solvers.ns2d.output"

        classes.PrintStdOut.module_name = base_name_mod + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutNS2D"

        classes.PhysFields.class_name = "PhysFieldsBase2D"

        classes._set_child(
            "Spectra",
            attribs={
                "module_name": base_name_mod + ".spectra",
                "class_name": "SpectraNS2D",
            },
        )

        classes._set_child(
            "SpectraMultiDim",
            attribs={
                "module_name": base_name_mod + ".spectra_multidim",
                "class_name": "SpectraMultiDimNS2D",
            },
        )

        classes._set_child(
            "SpatialMeans",
            attribs={
                "module_name": base_name_mod + ".spatial_means",
                "class_name": "SpatialMeansNS2D",
            },
        )

        classes._set_child(
            "SpectEnergyBudg",
            attribs={
                "module_name": base_name_mod + ".spect_energy_budget",
                "class_name": "SpectralEnergyBudgetNS2D",
            },
        )

        classes._set_child(
            "Increments",
            attribs={
                "module_name": "fluidsim.base.output.increments",
                "class_name": "Increments",
            },
        )

        classes._set_child(
            "TemporalSpectra",
            attribs={
                "module_name": "fluidsim.base.output.temporal_spectra",
                "class_name": "TemporalSpectra2D",
            },
        )

        classes._set_child(
            "SpatioTemporalSpectra",
            attribs={
                "module_name": base_name_mod + ".spatiotemporal_spectra",
                "class_name": "SpatioTemporalSpectraNS2D",
            },
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """Complete the `params` container (static method)."""
        OutputBasePseudoSpectral._complete_params_with_default(
            params, info_solver
        )

        params.output.phys_fields.field_to_plot = "rot"

    def compute_energy_fft(self):
        """Compute energy(k)"""
        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
        return (np.abs(ux_fft) ** 2 + np.abs(uy_fft) ** 2) / 2.0

    def compute_enstrophy_fft(self):
        """Compute enstrophy(k)"""
        rot_fft = self.sim.state.state_spect.get_var("rot_fft")
        return np.abs(rot_fft) ** 2 / 2.0

    def compute_energy(self):
        """Compute the spatially averaged energy."""
        energy_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(energy_fft)

    def compute_enstrophy(self):
        """Compute the spatially averaged enstrophy."""
        enstrophy_fft = self.compute_enstrophy_fft()
        return self.sum_wavenumbers(enstrophy_fft)



"""Normal mode decomposition for SW1L solvers
(:mod:`fluidsim.solvers.sw1l.output.normal_mode`)
=================================================

Provides:

.. autoclass:: NormalModeBase
   :members:
   :private-members:

.. autoclass:: NormalModeDecomposition
   :members:
   :private-members:

.. autoclass:: NormalModeDecompositionModified
   :members:
   :private-members:

"""

import numpy as np

from fluiddyn.util.compat import cached_property
from fluiddyn.util import mpi


class NormalModeBase:
    def __init__(self, output):
        self.sim = output.sim
        self.params = output.sim.params
        self.oper = output.oper

        f = self.params.f
        c = self.params.c2 ** 0.5
        ck = c * self.oper.K_not0

        if f == 0:
            self.sigma = ck
        else:
            self.sigma = np.sqrt(f ** 2 + (ck) ** 2)

        self.bvec_fft = None
        self.it_bvec_fft_computed = -1

    def compute(self):
        if self.it_bvec_fft_computed != self.sim.time_stepping.it:
            get_var = self.sim.state.get_var
            if {"ap_fft", "am_fft"}.issubset(self.sim.state.keys_state_spect):
                q_fft = get_var("q_fft")
                ap_fft = get_var("ap_fft")
                am_fft = get_var("am_fft")
                bvec_fft = self.bvecfft_from_qapamfft(q_fft, ap_fft, am_fft)
            else:
                ux_fft = get_var("ux_fft")
                uy_fft = get_var("uy_fft")
                eta_fft = get_var("eta_fft")
                bvec_fft = self.bvecfft_from_uxuyetafft(ux_fft, uy_fft, eta_fft)

            self.bvec_fft = bvec_fft
            self.it_bvec_fft_computed = self.sim.time_stepping.it

        return self.bvec_fft

    def bvecfft_from_qapamfft(self, q_fft, ap_fft, am_fft):
        r"""Compute normal mode vector :math:`\mathbf{B}`
        with dimensions of velocity from diagonalized linear modes."""
        c = self.params.c2 ** 0.5
        c2 = self.params.c2
        K = self.oper.K_not0
        sigma = self.sigma

        q_fft = -q_fft * c / sigma
        ap_fft = ap_fft * 2 ** 0.5 * c2 / (sigma * K)
        am_fft = am_fft * 2 ** 0.5 * c2 / (sigma * K)
        bvec_fft = np.array([q_fft, ap_fft, am_fft])
        if mpi.rank == 0 or self.oper.is_sequential:
            bvec_fft[:, 0, 0] = 0.0

        return bvec_fft

    def bvecfft_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft):
        r"""Compute normal mode vector, :math:`\mathbf{B}`
        with dimensions of velocity from primitive variables."""
        q_fft, ap_fft, am_fft = self.oper.qapamfft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        return self.bvecfft_from_qapamfft(q_fft, ap_fft, am_fft)

    def compute_qda_energies_fft(self):
        """Compute quadratic geostrophic, divergent and ageostrophic energies."""
        bvec_fft = self.compute()
        q_fft = bvec_fft[0]
        d_fft = 0.5 * (bvec_fft[1] - bvec_fft[2])
        a_fft = 0.5 * (bvec_fft[1] + bvec_fft[2])

        def energy(var_fft):
            return 0.5 * np.abs(var_fft) ** 2

        return energy(q_fft), energy(d_fft), energy(a_fft)

    def compute_qapam_energies_fft(self):
        """Compute quadratic geostrophic, and ageostrophic (+/-) energies."""
        bvec_fft = self.compute()

        def energy(var_fft):
            return 0.5 * np.abs(var_fft) ** 2

        return map(energy, bvec_fft)


class NormalModeDecomposition(NormalModeBase):
    def __init__(self, output):
        super().__init__(output)

    @cached_property
    def qmat(self):
        """Compute Q matrix to transform q, ap, am (fft) -> b0, b+, b- (fft) with
        dimensions of velocity.

        """
        oper = self.oper
        sigma = self.sigma

        f = float(self.params.f)
        c = self.params.c2 ** 0.5
        KX = oper.KX
        KY = oper.KY
        K = oper.K
        K2 = oper.K2
        K_not0 = oper.K_not0
        ck = c * K_not0

        qmat = (
            np.array(
                [
                    [
                        -1j * 2.0 ** 0.5 * ck * KY,
                        +1j * f * KY + KX * sigma,
                        +1j * f * KY - KX * sigma,
                    ],
                    [
                        +1j * 2.0 ** 0.5 * ck * KX,
                        -1j * f * KX + KY * sigma,
                        -1j * f * KX - KY * sigma,
                    ],
                    [2.0 ** 0.5 * f * K, c * K2, c * K2],
                ]
            )
            / (2.0 ** 0.5 * sigma * K_not0)
        )

        if mpi.rank == 0 or oper.is_sequential:
            qmat[:, :, 0, 0] = 0.0

        return qmat

    def normalmodefft_from_keyfft(self, key):
        """Returns the normal mode decomposition for the state_spect key specified."""

        if key == "div_fft":
            key_modes, normal_mode_vec_fft_x = self.normalmodefft_from_keyfft(
                "px_ux_fft"
            )
            key_modes, normal_mode_vec_fft_y = self.normalmodefft_from_keyfft(
                "py_uy_fft"
            )
            normal_mode_vec_fft = normal_mode_vec_fft_x + normal_mode_vec_fft_y
        else:
            key_modes = np.array([["G", "A", "a"]])
            row_index = {
                "ux_fft": 0,
                "uy_fft": 1,
                "eta_fft": 2,
                "px_ux_fft": 0,
                "px_uy_fft": 1,
                "px_eta_fft": 2,
                "py_ux_fft": 0,
                "py_uy_fft": 1,
                "py_eta_fft": 2,
            }

            r = row_index[key]
            normal_mode_vec_fft = np.einsum(
                "i...,i...->i...", self.qmat[r], self.bvec_fft
            )
            if "px" in key:
                for r in range(3):
                    normal_mode_vec_fft[r] = self.oper.pxffft_from_fft(
                        normal_mode_vec_fft[r]
                    )
            elif "py" in key:
                for r in range(3):
                    normal_mode_vec_fft[r] = self.oper.pyffft_from_fft(
                        normal_mode_vec_fft[r]
                    )

            if "eta" in key:
                normal_mode_vec_fft = normal_mode_vec_fft / self.params.c2 ** 0.5

        return key_modes, normal_mode_vec_fft

    def normalmodephys_from_keyphys(self, key):
        ifft2 = self.oper.ifft2
        key_modes, normal_mode_vec_fft = self.normalmodefft_from_keyfft(
            key + "_fft"
        )
        normal_mode_vec_phys = np.array(
            [ifft2(normal_mode_vec_fft[i]) for i in range(3)]
        )

        return key_modes, normal_mode_vec_phys

    def _group_matrix_using_dict(self, key_matrix, value_matrix, grouping):
        value_dict = dict.fromkeys(grouping, 0.0)
        n1, n2 = key_matrix.shape
        for i in range(n1):
            for j in range(n2):
                k1 = key_matrix[i, j]
                k3 = None
                for k2 in grouping.keys():
                    if k1 in grouping[k2]:
                        k3 = k2
                        break

                if k3 is None:
                    raise KeyError(
                        "Not sure which dyad group " + k1 + " belongs to"
                    )

                value_dict[k3] += value_matrix[i, j]

        new_matrix = np.array([value_dict[k] for k in value_dict.keys()])
        new_keys = np.array([list(value_dict)])
        return new_keys, new_matrix

    def dyad_from_keyfft(self, conjugate=False, *keys_state_spect):
        dyad_group = {
            "GG": ["GG"],
            "AG": ["GA", "AG"],
            "aG": ["Ga", "aG"],
            "AA": ["AA", "Aa", "aA", "aa"],
        }
        k1, k2 = keys_state_spect

        normal_modes = dict()
        if k1 != k2:
            for k in keys_state_spect:
                key_modes, normal_modes[k] = self.normalmodefft_from_keyfft(k)
        else:
            key_modes, normal_modes[k1] = self.normalmodefft_from_keyfft(k1)
            normal_modes[k2] = normal_modes[k1]

        key_modes_mat = np.char.add(key_modes.transpose(), key_modes)
        if conjugate:
            Ni = normal_modes[k1].conj()
            Nj = normal_modes[k2]
        else:
            Ni = normal_modes[k1]
            Nj = normal_modes[k2]
        dyad_mat_fft = np.einsum("i...,j...->ij...", Ni, Nj)
        del (normal_modes, Ni, Nj)
        return self._group_matrix_using_dict(
            key_modes_mat, dyad_mat_fft, dyad_group
        )

    def dyad_from_keyphys(self, *keys_state_phys):
        dyad_group = {
            "GG": ["GG"],
            "AG": ["GA", "AG"],
            "aG": ["Ga", "aG"],
            "AA": ["AA", "Aa", "aA", "aa"],
        }
        k1, k2 = keys_state_phys

        normal_modes = dict()
        if k1 != k2:
            for k in keys_state_phys:
                key_modes, normal_modes[k] = self.normalmodephys_from_keyphys(k)
        else:
            key_modes, normal_modes[k1] = self.normalmodephys_from_keyphys(k1)
            normal_modes[k2] = normal_modes[k1]
        key_modes_mat = np.char.add(key_modes.transpose(), key_modes)
        dyad_mat_phys = np.einsum(
            "i...,j...->ij...", normal_modes[k1], normal_modes[k2]
        )
        del normal_modes
        fft2 = self.oper.fft2
        dyad_mat_fft = np.array(
            [[fft2(dyad_mat_phys[i, j]) for j in range(3)] for i in range(3)]
        )

        for i in range(3):
            for j in range(3):
                self.oper.dealiasing(dyad_mat_fft[i, j])

        del dyad_mat_phys
        return self._group_matrix_using_dict(
            key_modes_mat, dyad_mat_fft, dyad_group
        )

    def triad_from_keyfft(self, *keys_state_spect):
        triad_group = {
            "GGG": ["GGG"],
            "AGG": ["AGG", "GAG", "GGA", "aGG", "GaG", "GGa"],
            "GAAs": ["aaG", "aGa", "Gaa", "AAG", "AGA", "GAA"],
            "GAAd": ["aAG", "AaG", "aGA", "AGa", "GaA", "GAa"],
            "AAA": ["AAA", "aaa", "AAa", "AaA", "aAA", "aaA", "aAa", "Aaa"],
        }
        k1, k2, k3 = keys_state_spect

        key_modes_1, normal_modes_1 = self.normalmodefft_from_keyfft(k1)
        key_modes_23, normal_modes_23 = self.dyad_from_keyfft(False, k2, k3)

        key_modes_mat = np.char.add(key_modes_1.transpose(), key_modes_23)
        triad_mat = np.einsum(
            "i...,j...->ij...", normal_modes_1.conj(), normal_modes_23
        )

        return self._group_matrix_using_dict(
            key_modes_mat, triad_mat, triad_group
        )

    def triad_from_keyfftphys(self, key_state_spect, *keys_state_phys):
        triad_group = {
            "GGG": ["GGG"],
            "AGG": ["AGG", "GAG", "GGA", "aGG", "GaG", "GGa"],
            "GAAs": ["aaG", "aGa", "Gaa", "AAG", "AGA", "GAA"],
            "GAAd": ["aAG", "AaG", "aGA", "AGa", "GaA", "GAa"],
            "AAA": ["AAA", "aaa", "AAa", "AaA", "aAA", "aaA", "aAa", "Aaa"],
        }
        k1 = key_state_spect
        k2, k3 = keys_state_phys

        key_modes_1, normal_modes_1 = self.normalmodefft_from_keyfft(k1)
        key_modes_23, normal_modes_23 = self.dyad_from_keyphys(k2, k3)

        key_modes_mat = np.char.add(key_modes_1.transpose(), key_modes_23)
        triad_mat = np.einsum(
            "i...,j...->ij...", normal_modes_1.conj(), normal_modes_23
        )

        return self._group_matrix_using_dict(
            key_modes_mat, triad_mat, triad_group
        )


class NormalModeDecompositionModified(NormalModeDecomposition):
    def compute(self):
        if self.it_bvec_fft_computed != self.sim.time_stepping.it:
            get_var = self.sim.state.get_var
            #
            # FIXME:Does not work
            #
            # if {'ap_fft', 'am_fft'}.issubset(self.sim.state.keys_state_spect):
            #     q_fft = get_var('q_fft')
            #     ap_fft = get_var('ap_fft')
            #     am_fft = get_var('am_fft')
            #     a_fft = ap_fft + am_fft
            #     bvecrot_fft = self.bvecfft_from_qapamfft(q_fft, a_fft, a_fft)
            # else:
            rot_fft = self.sim.state.get_var("rot_fft")
            uxr_fft, uyr_fft = self.oper.vecfft_from_rotfft(rot_fft)
            eta_fft = get_var("eta_fft")
            bvecrot_fft = self.bvecfft_from_uxuyetafft(uxr_fft, uyr_fft, eta_fft)

            self.bvecrot_fft = bvecrot_fft

        return super().compute()

    def normalmodefft_from_keyfft(self, key):
        """Returns the normal mode decomposition for the state_spect key specified."""
        if key.endswith("urx_fft") or key.endswith("ury_fft"):
            row_index = {
                "urx_fft": 0,
                "ury_fft": 1,
                "px_urx_fft": 0,
                "px_ury_fft": 1,
                "py_urx_fft": 0,
                "py_ury_fft": 1,
            }

            key_modes = np.array([["G", "A", "a"]])
            r = row_index[key]
            normal_mode_vec_fft = np.einsum(
                "i...,i...->i...", self.qmat[r], self.bvecrot_fft
            )
            if "px" in key:
                for r in range(3):
                    normal_mode_vec_fft[r] = self.oper.pxffft_from_fft(
                        normal_mode_vec_fft[r]
                    )
            elif "py" in key:
                for r in range(3):
                    normal_mode_vec_fft[r] = self.oper.pyffft_from_fft(
                        normal_mode_vec_fft[r]
                    )

            return key_modes, normal_mode_vec_fft

        else:
            return super().normalmodefft_from_keyfft(key)



"""NS2D over a sphere (:mod:`fluidsim.solvers.sphere.ns2d`)
===========================================================

Solving the incompressible Navier-Stokes equations over a sphere.

.. autosummary::
   :toctree:

   solver

"""



"""
Correl freq (:mod:`fluidsim.solvers.plate2d.output.correlations_freq`)
============================================================================


Provides:

.. autoclass:: CorrelationsFreq
   :members:
   :private-members:

"""

import os

import numpy as np
import h5py
import matplotlib.pyplot as plt

from transonic import jit
from fluiddyn.util import mpi
from fluiddyn.calcul.easypyfft import FFTW1DReal2Complex

from fluidsim.base.output.base import SpecificOutput


@jit
def compute_correl4_seq(
    q_fftt: "complex128[][]", iomegas1: "int32[]", nb_omegas: int, nb_xs_seq: int
):
    r"""Compute the correlations 4.

    .. math::
       C_4(\omega_1, \omega_2, \omega_3, \omega_4) =
       \langle
       \tilde w(\omega_1, \mathbf{x})
       \tilde w(\omega_2, \mathbf{x})
       \tilde w(\omega_3, \mathbf{x})^*
       \tilde w(\omega_4, \mathbf{x})^*
       \rangle_\mathbf{x},

    where

    .. math::
       \omega_2 = \omega_3 + \omega_4 - \omega_1

    and :math:`\omega_1 > 0`, :math:`\omega_3 > 0` and
    :math:`\omega_4 > 0`. Thus, this function produces an array
    :math:`C_4(\omega_1, \omega_3, \omega_4)`.

    """
    q_fftt_conj = np.conj(q_fftt)

    nx = q_fftt.shape[0]
    n0 = iomegas1.shape[0]

    corr4 = np.zeros((len(iomegas1), nb_omegas, nb_omegas), dtype=np.complex128)

    for i1 in range(n0):
        io1 = iomegas1[i1]
        # this loop could be parallelized (OMP)
        for io3 in range(nb_omegas):
            # we use the symmetry omega_3 <--> omega_4
            for io4 in range(io3 + 1):
                io2 = io3 + io4 - io1
                if io2 < 0:
                    io2 = -io2
                    for ix in range(nx):
                        corr4[i1, io3, io4] += (
                            q_fftt[ix, io1]
                            * q_fftt_conj[ix, io3]
                            * q_fftt_conj[ix, io4]
                            * q_fftt_conj[ix, io2]
                        )
                elif io2 >= nb_omegas:
                    io2 = 2 * nb_omegas - 1 - io2
                    for ix in range(nx):
                        corr4[i1, io3, io4] += (
                            q_fftt[ix, io1]
                            * q_fftt_conj[ix, io3]
                            * q_fftt_conj[ix, io4]
                            * q_fftt_conj[ix, io2]
                        )
                else:
                    for ix in range(nx):
                        corr4[i1, io3, io4] += (
                            q_fftt[ix, io1]
                            * q_fftt_conj[ix, io3]
                            * q_fftt_conj[ix, io4]
                            * q_fftt[ix, io2]
                        )
                # symmetry omega_3 <--> omega_4:
                corr4[i1, io4, io3] = corr4[i1, io3, io4]
    return corr4


@jit
def compute_correl2_seq(
    q_fftt: "complex128[][]", iomegas1: "int32[]", nb_omegas: int, nb_xs_seq: int
):
    r"""Compute the correlations 2.

    .. math::
       C_2(\omega_1, \omega_2) =
       \langle
       \tilde w(\omega_1, \mathbf{x})
       \tilde w(\omega_2, \mathbf{x})^*
       \rangle_\mathbf{x},

    where :math:`\omega_1 = \omega_2`. Thus, this function
    produces an array :math:`C_2(\omega)`.

    """
    q_fftt_conj = np.conj(q_fftt)
    nx = q_fftt.shape[0]

    corr2 = np.zeros((nb_omegas, nb_omegas), dtype=np.complex128)

    for io3 in range(nb_omegas):
        for io4 in range(io3 + 1):
            for ix in range(nx):
                corr2[io3, io4] += q_fftt[ix, io3] * q_fftt_conj[ix, io4]
            corr2[io4, io3] = np.conj(corr2[io3, io4])

    return corr2


def compute_correl4(q_fftt, iomegas1, nb_omegas, nb_xs_seq):
    corr4 = compute_correl4_seq(q_fftt, iomegas1, nb_omegas, nb_xs_seq)
    if mpi.nb_proc > 1:
        # reduce SUM for mean:
        corr4 = mpi.comm.reduce(corr4, op=mpi.MPI.SUM, root=0)

    if mpi.rank == 0:
        corr4 /= nb_xs_seq
        return corr4


def compute_correl2(q_fftt, iomegas1, nb_omegas, nb_xs_seq):
    corr2 = compute_correl2_seq(q_fftt, iomegas1, nb_omegas, nb_xs_seq)
    if mpi.nb_proc > 1:
        # reduce SUM for mean:
        corr2 = mpi.comm.reduce(corr2, op=mpi.MPI.SUM, root=0)

    if mpi.rank == 0:
        corr2 /= nb_xs_seq
        return corr2


class CorrelationsFreq(SpecificOutput):
    """Compute, save, load and plot correlations of frequencies."""

    _tag = "correl_freq"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "correl_freq"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(
            tag,
            attribs={
                "HAS_TO_PLOT_SAVED": False,
                "it_start": 10,
                "nb_times_compute": 100,
                "coef_decimate": 10,
                "key_quantity": "w",
                "iomegas1": [1],
            },
        )

    def __init__(self, output):
        params = output.sim.params
        pcorrel_freq = params.output.correl_freq
        super().__init__(
            output,
            period_save=params.output.periods_save.correl_freq,
            has_to_plot_saved=pcorrel_freq.HAS_TO_PLOT_SAVED,
        )

        self.nb_times_compute = pcorrel_freq.nb_times_compute
        self.coef_decimate = pcorrel_freq.coef_decimate
        self.key_quantity = pcorrel_freq.key_quantity
        self.iomegas1 = np.array(pcorrel_freq.iomegas1, dtype=np.int32)
        self.it_last_run = pcorrel_freq.it_start
        n0 = len(
            list(range(0, output.sim.oper.shapeX_loc[0], self.coef_decimate))
        )
        n1 = len(
            list(range(0, output.sim.oper.shapeX_loc[1], self.coef_decimate))
        )
        nb_xs = n0 * n1

        self.spatio_temp = np.empty([nb_xs, self.nb_times_compute])
        self.oper_fft1 = FFTW1DReal2Complex(self.spatio_temp.shape)
        self.nb_omegas = self.oper_fft1.shapeK[-1]
        self.hamming = np.hanning(self.nb_times_compute)

        if mpi.nb_proc > 1:
            nb_xs = mpi.comm.reduce(nb_xs, op=mpi.MPI.SUM, root=0)
        if mpi.rank > 0:
            nb_xs = 0
        self.nb_xs_seq = nb_xs

        self.nb_times_in_spatio_temp = 0

        if os.path.exists(self.path_file):
            with h5py.File(self.path_file, "r") as file:
                link_corr4 = file["corr4"]
                link_corr2 = file["corr2"]
                link_nb_means = file["nb_means"]
                self.corr4 = link_corr4[-1]
                self.corr2 = link_corr2[-1]
                self.nb_means_times = link_nb_means[-1]
                self.periods_fill = file["periods_fill"][...]
                if self.sim.time_stepping.deltat != file["deltat"][...]:
                    raise ValueError()

        else:
            self.periods_fill = params.output.periods_save.correl_freq
            self.corr4 = np.zeros(
                [len(self.iomegas1), self.nb_omegas, self.nb_omegas],
                dtype=np.complex128,
            )
            self.corr2 = np.zeros(
                [self.nb_omegas, self.nb_omegas], dtype=np.complex128
            )
            self.nb_means_times = 0

        if self.periods_fill > 0:
            dt_output = self.periods_fill * output.sim.time_stepping.deltat
            duration = self.nb_times_compute * dt_output
            delta_omega = 2 * np.pi / duration
            self.omegas = delta_omega * np.arange(self.nb_omegas)

            self.omega_Nyquist = np.pi / dt_output

            omega1_max = self.iomegas1.max() * delta_omega
            if self.omega_Nyquist <= omega1_max:
                raise ValueError(
                    f"omega_1_max={omega1_max} is larger than "
                    f"omega_Nyquist={self.omega_Nyquist}."
                )

            self.omega_dealiasing = (
                self.params.oper.coef_dealiasing
                * np.pi
                * self.params.oper.nx
                / self.params.oper.Lx
            ) ** 2

            if self.omega_dealiasing > self.omega_Nyquist:
                print("Warning: omega_dealiasing > omega_Nyquist")

    def _init_files(self, arrays_1st_time=None):
        # we can not do anything when this function is called.
        pass

    def _init_files2(self, correlations):
        time_tot = (
            self.sim.time_stepping.deltat
            * self.nb_times_compute
            * self.periods_fill
        )
        omegas = 2 * np.pi / time_tot * np.arange(self.nb_omegas)
        arrays_1st_time = {
            "omegas": omegas,
            "deltat": self.sim.time_stepping.deltat,
            "nb_times_compute": self.nb_times_compute,
            "periods_fill": self.periods_fill,
        }
        self._create_file_from_dict_arrays(
            self.path_file, correlations, arrays_1st_time
        )

        self.t_last_save = self.sim.time_stepping.t

    def _online_save(self):
        """Save the values at one time. """
        itsim = self.sim.time_stepping.t / self.sim.time_stepping.deltat
        periods_save = self.sim.params.output.periods_save.correl_freq

        if itsim - self.it_last_run >= periods_save - 1:
            self.it_last_run = itsim
            field = self.sim.state.state_phys.get_var(self.key_quantity)
            field = field[:: self.coef_decimate, :: self.coef_decimate]
            self.spatio_temp[:, self.nb_times_in_spatio_temp] = field.reshape(
                [field.size]
            )
            self.nb_times_in_spatio_temp += 1
            if self.nb_times_in_spatio_temp == self.nb_times_compute:
                self.nb_times_in_spatio_temp = 0
                self.t_last_save = self.sim.time_stepping.t
                spatio_fft = self.oper_fft1.fft(self.hamming * self.spatio_temp)
                new_corr4 = compute_correl4(
                    spatio_fft, self.iomegas1, self.nb_omegas, self.nb_xs_seq
                )
                new_corr2 = compute_correl2(
                    spatio_fft, self.iomegas1, self.nb_omegas, self.nb_xs_seq
                )

                if mpi.rank == 0:
                    self.corr4 = (1.0 / (self.nb_means_times + 1)) * (
                        self.nb_means_times * self.corr4 + new_corr4
                    )
                    self.corr2 = (1.0 / (self.nb_means_times + 1)) * (
                        self.nb_means_times * self.corr2 + new_corr2
                    )
                    self.nb_means_times += 1

                    if (
                        self.nb_means_times % 128 == 0
                        or np.log(self.nb_means_times) / np.log(2) % 1 == 0
                    ) and self.nb_means_times != 1:

                        correlations = {
                            "corr4": self.corr4,
                            "corr2": self.corr2,
                            "nb_means": self.nb_means_times,
                        }
                        if not os.path.exists(self.path_file):
                            self._init_files2(correlations)
                        else:
                            # save the spectra in the file correl_freq.h5
                            self._add_dict_arrays_to_file(
                                self.path_file, correlations
                            )
                        if self.has_to_plot:
                            self._online_plot_saving(correlations)

    #     if (tsim-self.t_last_show >= self.period_show):
    #         self.t_last_show = tsim
    #         self.ax.get_figure().canvas.draw()

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, ax = self.output.figure_axe(numfig=4_100_000)
            self.ax = ax
            ax.set_xlabel("Omega")
            ax.set_ylabel("Correlations")
            ax.set_title("Correlation\n" + self.output.summary_simul)

    def _online_plot_saving(self, dict_results):
        nb_omegas = self.nb_omegas

        corr4 = dict_results["corr4"]
        corr2 = dict_results["corr2"]
        corr = np.empty(corr4.shape)
        fy, fx = np.meshgrid(self.omegas, self.omegas)
        for i1, io1 in enumerate(self.iomegas1):
            for io3 in range(nb_omegas):
                for io4 in range(io3 + 1):
                    io2 = io3 + io4 - io1
                    if io2 < 0:
                        io2 = -io2
                    elif io2 >= nb_omegas:
                        io2 = 2 * nb_omegas - 1 - io2
                    corr[i1, io3, io4] = abs(corr4[i1, io3, io4]) / np.sqrt(
                        abs(
                            corr2[io1, io1]
                            * corr2[io3, io3]
                            * corr2[io4, io4]
                            * corr2[io2, io2]
                        )
                    )
                    corr[i1, io4, io3] = corr[i1, io3, io4]

        fig, ax = self.output.figure_axe(numfig=4_100_000)
        self.ax = ax
        ax.set_xlabel("Omega")
        ax.set_xlabel("Correlation")
        ax.plot(corr2[:, :], "k.")

        fig, ax = self.output.figure_axe(numfig=2333)
        self.ax = ax
        ax.set_xlabel("Omega")
        ax.set_ylabel("Omega")
        pc = ax.pcolormesh(fx, fy, abs(corr[0, :, :]), shading="nearest")
        fig.colorbar(pc)

        fig1, ax1 = self.output.figure_axe(numfig=2334)
        self.ax1 = ax1
        ax1.set_xlabel("Omega")
        ax1.set_ylabel("Omega")
        pc1 = ax1.pcolormesh(fx, fy, corr[1, :, :], shading="nearest")
        fig1.colorbar(pc1)

    def compute_corr4_norm(self, it=-1):

        with h5py.File(self.path_file, "r") as file:
            corr4 = file["corr4"][it]
            corr2 = file["corr2"][it]
            nb_means = file["nb_means"][it]

        nb_omegas = self.nb_omegas

        corr_norm = np.empty_like(corr4)
        cum_norm = np.empty(corr4.shape)
        norm = np.empty(corr4.shape)
        tmp1 = np.empty((nb_omegas, nb_omegas), dtype=np.complex128)
        tmp2 = np.empty((nb_omegas, nb_omegas), dtype=np.complex128)
        for i1, io1 in enumerate(self.iomegas1):
            for io3 in range(nb_omegas):
                for io4 in range(io3 + 1):
                    io2 = io3 + io4 - io1
                    if io2 < 0:
                        io2 = -io2
                    elif io2 >= nb_omegas:
                        io2 = 2 * nb_omegas - 1 - io2

                    norm[i1, io3, io4] = np.sqrt(
                        abs(
                            corr2[io1, io1]
                            * corr2[io3, io3]
                            * corr2[io4, io4]
                            * corr2[io2, io2]
                        )
                    )
                    norm[i1, io4, io3] = norm[i1, io3, io4]

                    tmp1[io3, io4] = corr2[io4, io2].conj() * corr2[io1, io3]
                    tmp2[io3, io4] = corr2[io1, io4] * corr2[io3, io2].conj()

                    cum_norm[i1, io3, io4] = (
                        abs(corr4[i1, io3, io4] - tmp1[io3, io4] - tmp2[io3, io4])
                        / norm[i1, io3, io4]
                    )
                    cum_norm[i1, io4, io3] = cum_norm[i1, io3, io4]

                    corr_norm[i1, io3, io4] = abs(
                        corr4[i1, io3, io4] / norm[i1, io3, io4]
                    )
                    corr_norm[i1, io4, io3] = corr_norm[i1, io3, io4]

        return norm, corr_norm, cum_norm, nb_means

    # def _compute_correl4(self, q_fftt):
    #     r"""Compute the correlations 4.

    #     .. math::
    #        C_4(\omega_1, \omega_2, \omega_3, \omega_4) =
    #        \langle
    #        \tilde w(\omega_1, \mathbf{x})
    #        \tilde w(\omega_2, \mathbf{x})
    #        \tilde w(\omega_3, \mathbf{x})^*
    #        \tilde w(\omega_4, \mathbf{x})^*
    #        \rangle_\mathbf{x},

    #     where

    #     .. math::
    #        \omega_2 = \omega_3 + \omega_4 - \omega_1

    #     and :math:`\omega_1 > 0`, :math:`\omega_3 > 0` and

    #     :math:`\omega_4 > 0`. Thus, this function produces an array
    #     :math:`C_4(\omega_1, \omega_3, \omega_4)`.

    #     """

    #     q_fftt_conj = q_fftt.conj()
    #     nb_omegas = self.nb_omegas

    #     corr4 = np.empty([len(self.iomegas1), nb_omegas, nb_omegas])
    #     for i1, io1 in enumerate(self.iomegas1):
    #         # this loop could be parallelized (OMP)
    #         for io3 in range(nb_omegas):
    #             # we use the symmetry omega_3 <--> omega_4
    #             for io4 in range(io3 + 1):
    #                 tmp = (
    #                     q_fftt[:, io1] * q_fftt_conj[:, io3] * q_fftt_conj[:, io4]
    #                 )
    #                 io2 = io3 + io4 - io1
    #                 if io2 < 0:
    #                     io2 = -io2
    #                     corr4[i1, io3, io4] = np.sum(tmp * q_fftt_conj[:, io2])
    #                 elif io2 >= nb_omegas:
    #                     io2 = 2 * nb_omegas - 1 - io2
    #                     corr4[i1, io3, io4] = np.sum(tmp * q_fftt_conj[:, io2])
    #                 else:
    #                     corr4[i1, io3, io4] = np.sum(tmp * q_fftt[:, io2])
    #                 # symmetry omega_3 <--> omega_4:
    #                 corr4[i1, io4, io3] = corr4[i1, io3, io4]

    #     if mpi.nb_proc > 1:
    #         # reduce SUM for mean:
    #         corr4 = mpi.comm.reduce(corr4, op=mpi.MPI.SUM, root=0)

    #     if mpi.rank == 0:
    #         corr4 /= self.nb_xs_seq
    #         return corr4

    # def _compute_correl2(self, q_fftt):
    #     r"""Compute the correlations 2.

    #     .. math::
    #        C_2(\omega_1, \omega_2) =
    #        \langle
    #        \tilde w(\omega_1, \mathbf{x})
    #        \tilde w(\omega_2, \mathbf{x})^*
    #        \rangle_\mathbf{x}.

    #     """
    #     nb_omegas = self.nb_omegas
    #     corr2 = np.empty([nb_omegas, nb_omegas])

    #     q_fftt_conj = q_fftt.conj()
    #     for io3 in range(nb_omegas):
    #         for io4 in range(io3 + 1):
    #             corr2[io3, io4] = np.sum(q_fftt[:, io3] * q_fftt_conj[:, io4])
    #             corr2[io4, io3] = corr2[io3, io4].conj()

    #     if mpi.nb_proc > 1:
    #         # reduce SUM for mean:
    #         corr2 = mpi.comm.reduce(corr2, op=mpi.MPI.SUM, root=0)

    #     if mpi.rank == 0:
    #         corr2 /= self.nb_xs_seq
    #         return corr2

    def _compute_norm_pick_corr4_from_corr4(self, corr4, i1=0, delta_io=10):
        io1 = self.iomegas1[i1]
        io3 = 4 * io1
        io4 = 4 * io1
        delta_io = min(delta_io, io3)
        if corr4.ndim == 3:
            return np.abs(
                np.mean(
                    corr4[
                        i1,
                        io3 - delta_io : io3 + delta_io + 1,
                        io4 - delta_io : io4 + delta_io + 1,
                    ]
                )
            )

        elif corr4.ndim == 4:
            corr4_mini = corr4[
                :,
                i1,
                io3 - delta_io : io3 + delta_io + 1,
                io4 - delta_io : io4 + delta_io + 1,
            ]
            return np.abs(corr4_mini.mean((1, 2)))

    def _compute_norm_pick_corr4(self):
        with h5py.File(self.path_file, "r") as file:
            corr4 = file["corr4"][:]
            nb_means = file["nb_means"][:]

        fcorr4 = self._compute_norm_pick_corr4_from_corr4(corr4)

        return nb_means, fcorr4

    def _compute_dnormpickC4_over_dnbmean(self):
        with h5py.File(self.path_file, "r") as file:
            corr4 = file["corr4"][:]
            nb_means = file["nb_means"][:]

        fcorr4 = self._compute_norm_pick_corr4_from_corr4(corr4)

        return (
            nb_means,
            np.absolute(
                np.diff(fcorr4) / np.diff(nb_means) * nb_means[1:] / fcorr4[1:]
            ),
        )

    def plot_norm_pick_corr4(self):
        nb_means, fcorr4 = self._compute_norm_pick_corr4()
        plt.figure()
        ax = plt.gca()
        ax.loglog(nb_means, fcorr4, "x-")
        ax.set_ylabel("a kind of norm of corr4")
        ax.set_xlabel("number of averages")

    def plot_convergence(self):

        nb_means, dnormpickC4 = self._compute_dnormpickC4_over_dnbmean()

        fig = plt.figure()
        fig.suptitle('"convergence"')
        ax = plt.gca()
        ax.loglog(nb_means[1:], dnormpickC4, "x-")
        ax.set_xlabel("number of averages")

    def plot_corr2(self, nonorm=False, it=-1):

        with h5py.File(self.path_file, "r") as file:
            corr2_in_file = file["corr2"]
            corr2 = corr2_in_file[it]
            nb_means = file["nb_means"][it]

        fy, fx = np.meshgrid(self.omegas, self.omegas)

        if nonorm:
            corr2_norm = corr2
        else:
            corr2_norm = np.empty(
                (self.nb_omegas, self.nb_omegas), dtype=np.complex128
            )
            for io3 in range(self.nb_omegas):
                for io4 in range(io3 + 1):
                    corr2_norm[io3, io4] = corr2[io3, io4] / np.sqrt(
                        np.absolute(corr2[io3, io3] * corr2[io4, io4])
                    )
                    corr2_norm[io4, io3] = corr2_norm[io3, io4].conj()

        log10corr2 = np.log10(abs(corr2_norm))
        plt.figure()
        ax = plt.gca()
        ax.set_title("log10(abs(corr2)); nb_means: " + str(nb_means))
        plt.xlabel("Omega")
        plt.ylabel("Omega")
        plt.pcolormesh(
            fx,
            fy,
            log10corr2,
            shading="nearest",
            vmin=log10corr2.min(),
            vmax=log10corr2.max(),
        )
        plt.colorbar()
        plt.axis([fx.min(), fx.max(), fy.min(), fy.max()])

    def plot_corr2_1d(self, it=-1):

        with h5py.File(self.path_file, "r") as file:
            corr2 = file["corr2"][it]
            nb_means = file["nb_means"][it]

        corr2_diag = np.empty(self.nb_omegas, dtype=np.complex128)
        for io3 in range(self.nb_omegas):
            corr2_diag[io3] = corr2[io3, io3]

        fig = plt.figure(num=18)
        fig.clf()
        ax = plt.gca()
        ax.set_title("abs(corr2_diag); nb_means: " + str(nb_means))
        plt.xlabel("Omega")
        plt.ylabel("abs(corr2)")
        # ax.loglog(self.omegas, abs(corr2_diag))
        ax.plot(self.omegas, np.log10(abs(corr2_diag)))

    def plot_corr4(self, it=-1):

        nb_omegas1 = self.iomegas1.shape[0]
        nb_omegas = self.nb_omegas

        corr_norm = np.empty((nb_omegas1, nb_omegas, nb_omegas))
        cum_norm = np.empty(corr_norm.shape)
        norm = np.empty(corr_norm.shape)
        norm, corr_norm, cum_norm, nb_means = self.compute_corr4_norm(it)

        fy, fx = np.meshgrid(self.omegas, self.omegas)

        fig = plt.figure()
        fig.suptitle("log10(abs(corr_norm)); nb_means: " + str(nb_means))
        nb_subplot_vert = int(np.sqrt(nb_omegas1))
        nb_subplot_horiz = int(round(nb_omegas1 / nb_subplot_vert))
        for i1, io1 in enumerate(self.iomegas1):
            plt.subplot(nb_subplot_vert, nb_subplot_horiz, i1 + 1)
            plt.xlabel("Omega")
            plt.ylabel("Omega")
            plt.pcolormesh(
                fx,
                fy,
                np.log10(abs(corr_norm[i1, :, :])),
                shading="nearest",
                vmin=np.log10(abs(corr_norm.min())),
                vmax=np.log10(abs(corr_norm.max())),
            )
            plt.colorbar()
            plt.axis([fx.min(), fx.max(), fy.min(), fy.max()])

        fig = plt.figure()
        fig.suptitle("log10(abs(cum_norm)); nb_means: " + str(nb_means))
        for i1, io1 in enumerate(self.iomegas1):
            plt.subplot(nb_subplot_vert, nb_subplot_horiz, i1 + 1)
            plt.xlabel("Omega")
            plt.ylabel("Omega")
            plt.pcolormesh(
                fx,
                fy,
                np.log10(abs(cum_norm[i1, :, :])),
                shading="nearest",
                vmin=np.log10(abs(cum_norm.min())),
                vmax=np.log10(abs(cum_norm.max())),
            )
            plt.colorbar()

            plt.axis([fx.min(), fx.max(), fy.min(), fy.max()])
        fig = plt.figure()
        fig.suptitle("log10(abs(norm)); nb_means: " + str(nb_means))
        for i1, io1 in enumerate(self.iomegas1):
            plt.subplot(nb_subplot_vert, nb_subplot_horiz, i1 + 1)
            plt.xlabel("Omega")
            plt.ylabel("Omega")
            plt.pcolormesh(
                fx,
                fy,
                np.log10(abs(norm[i1, :, :])),
                shading="nearest",
                vmin=np.log10(abs(norm.min())),
                vmax=np.log10(abs(norm.max())),
            )
            plt.colorbar()
            plt.axis([fx.min(), fx.max(), fy.min(), fy.max()])

    def plot_convergence2(self):
        ws = 10
        with h5py.File(self.path_file, "r") as file:
            corr4_full = file["corr4"][...]
            nb_means = file["nb_means"][...]

        means = np.zeros(corr4_full.shape[0])
        f_conv = np.zeros(corr4_full.shape[0:2])
        for inb in range(corr4_full.shape[0]):
            corr4_nb = corr4_full[inb]
            means[inb] = nb_means[inb]
            for i1, io1 in enumerate(self.iomegas1):
                print(io1, ws, self.nb_omegas)
                if (2 * io1 + ws < self.nb_omegas) and (2 * io1 - ws > io1):
                    f_conv[inb, i1] = np.abs(
                        np.mean(
                            corr4_nb[
                                i1,
                                2 * io1 - ws : 2 * io1 + ws + 1,
                                2 * io1 - ws : 2 * io1 + ws + 1,
                            ]
                        )
                    )
        fig, ax1 = self.output.figure_axe()
        ax1.set_xlabel("nb_means")
        # ax1.set_ylabel('convergence')
        ax1.set_title("Trispectra Convergence")

        if f_conv.max() == 0:
            return

        for i1, io1 in enumerate(self.iomegas1):
            norm = f_conv[-1, i1]
            if norm != 0:
                f_conv[:, i1] = f_conv[:, i1] / norm
            ax1.plot(means, f_conv[:, i1], linewidth=1)

        ax1.set_xscale("log")
        ax1.set_yscale("log")



import os
import sys
import unittest
from glob import glob
from shutil import rmtree

from fluiddyn.util import mpi
from fluidsim.util.testing import TestCase, skip_if_no_fluidfft

from .__main__ import run_profile

if mpi.rank == 0:
    pid = os.getpid()
else:
    pid = None

if mpi.nb_proc > 1:
    pid = mpi.comm.bcast(pid, root=0)

path_dir = f"/tmp/tmp_test_fluidsim_profile_dir_pid{pid}"


class TestsProfile(TestCase):
    @classmethod
    def tearDownClass(cls):
        if mpi.rank == 0 and os.path.isdir(path_dir):
            rmtree(path_dir)

    @skip_if_no_fluidfft
    def test3d(self):
        command = "fluidsim-profile 8 -s ns3d -o"
        args = command.split()
        args.append(path_dir)
        sys.argv = args
        run_profile()

        if mpi.nb_proc == 1:
            command = "fluidsim-profile -p -sf"
            args = command.split()
            paths = glob(path_dir + "/*")
            path = paths[0]
            args.append(path)
            sys.argv = args
            run_profile()

    @skip_if_no_fluidfft
    def test2d(self):
        command = "fluidsim-profile 16 -s ns2d -o"
        args = command.split()
        args.append(path_dir)
        sys.argv = args
        run_profile()

    @unittest.skipIf(
        mpi.nb_proc > 1, "Profiling solver ad1d is not meant to be done with MPI"
    )
    def test1d(self):
        command = "fluidsim-profile 128 -d1 -s ad1d -o"
        args = command.split()
        args.append(path_dir)
        sys.argv = args

        # FIXME: works with pytest, but not with unittest
        #  from fluidsim.util.console.util import ConsoleError
        #  with self.assertRaises(ConsoleError):
        #      # No profiling implemented for 1D solvers
        #      run_profile()


if __name__ == "__main__":
    unittest.main()
